# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Read/Write Avro File Object Containers.
"""
import zlib
import six
from six import BytesIO as StringIO
from spavro import schema
from spavro import io
try:
    import snappy
    has_snappy = True
except ImportError:
    has_snappy = False
#
# Constants
#

VERSION = 1
MAGIC = b'Obj' + six.int2byte(VERSION)  # chr(VERSION)
MAGIC_SIZE = len(MAGIC)
SYNC_SIZE = 16
SYNC_INTERVAL = 4000 * SYNC_SIZE # TODO(hammer): make configurable
META_SCHEMA = schema.parse("""\
{"type": "record", "name": "org.apache.avro.file.Header",
 "fields" : [
     {"name": "magic", "type": {"type": "fixed", "name": "magic", "size": %d}},
     {"name": "meta", "type": {"type": "map", "values": "bytes"}},
     {"name": "sync", "type": {"type": "fixed", "name": "sync", "size": %d}}]}
""" % (MAGIC_SIZE, SYNC_SIZE))
VALID_CODECS = ['null', 'deflate']
if has_snappy:
        VALID_CODECS.append('snappy')
VALID_ENCODINGS = ['binary'] # not used yet

CODEC_KEY = "avro.codec"
SCHEMA_KEY = "avro.schema"

#
# Exceptions
#


class DataFileException(schema.AvroException):
    """
    Raised when there's a problem reading or writing file object containers.
    """
    def __init__(self, fail_msg):
        schema.AvroException.__init__(self, fail_msg)

#
# Write Path
#


class DataFileWriter(object):
    @staticmethod
    def generate_sync_marker():
        return generate_sixteen_random_bytes()

    # TODO(hammer): make 'encoder' a metadata property
    def __init__(self, writer, datum_writer, writers_schema=None, codec='null'):
        """
        If the schema is not present, presume we're appending.

        @param writer: File-like object to write into.
        """
        self._writer = writer
        self._encoder = io.BinaryEncoder(writer)
        self._datum_writer = datum_writer
        self._buffer_writer = StringIO()
        self._buffer_encoder = io.BinaryEncoder(self._buffer_writer)
        self._block_count = 0
        self._meta = {}
        self._header_written = False

        if writers_schema is not None:
            if codec not in VALID_CODECS:
                raise DataFileException("Unknown codec: %r" % codec)
            self._sync_marker = DataFileWriter.generate_sync_marker()
            self.set_meta('avro.codec', codec)
            self.set_meta('avro.schema', str(writers_schema))
            self.datum_writer.writers_schema = writers_schema
        else:
            # open writer for reading to collect metadata
            dfr = DataFileReader(writer, io.DatumReader())
            
            # TODO(hammer): collect arbitrary metadata
            # collect metadata
            self._sync_marker = dfr.sync_marker
            self.set_meta('avro.codec', dfr.get_meta('avro.codec'))

            # get schema used to write existing file
            schema_from_file = dfr.get_meta('avro.schema')
            self.set_meta('avro.schema', schema_from_file)
            self.datum_writer.writers_schema = schema.parse(schema_from_file)

            # seek to the end of the file and prepare for writing
            writer.seek(0, 2)
            self._header_written = True

    # read-only properties
    writer = property(lambda self: self._writer)
    encoder = property(lambda self: self._encoder)
    datum_writer = property(lambda self: self._datum_writer)
    buffer_writer = property(lambda self: self._buffer_writer)
    buffer_encoder = property(lambda self: self._buffer_encoder)
    sync_marker = property(lambda self: self._sync_marker)
    meta = property(lambda self: self._meta)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # Perform a close if there's no exception
        if type is None:
            self.close()

    # read/write properties
    def set_block_count(self, new_val):
        self._block_count = new_val
    block_count = property(lambda self: self._block_count, set_block_count)

    # utility functions to read/write metadata entries
    def get_meta(self, key):
        return self._meta.get(key).decode('utf-8')
    def set_meta(self, key, val):
        self._meta[key] = val.encode('utf-8')

    def _write_header(self):
        header = {'magic': MAGIC,
                            'meta': self.meta,
                            'sync': self.sync_marker}
        self.datum_writer.write_data(META_SCHEMA, header, self.encoder)
        self._header_written = True

    # TODO(hammer): make a schema for blocks and use datum_writer
    def _write_block(self):
        if not self._header_written:
            self._write_header()

        if self.block_count > 0:
            # write number of items in block
            self.encoder.write_long(self.block_count)

            # write block contents
            uncompressed_data = self.buffer_writer.getvalue()
            if self.get_meta(CODEC_KEY) == 'null':
                compressed_data = uncompressed_data
                compressed_data_length = len(compressed_data)
            elif self.get_meta(CODEC_KEY) == 'deflate':
                # The first two characters and last character are zlib
                # wrappers around deflate data.
                compressed_data = zlib.compress(uncompressed_data)[2:-1]
                compressed_data_length = len(compressed_data)
            elif self.get_meta(CODEC_KEY) == 'snappy':
                compressed_data = snappy.compress(uncompressed_data)
                compressed_data_length = len(compressed_data) + 4 # crc32
            else:
                fail_msg = '"%s" codec is not supported.' % self.get_meta(CODEC_KEY)
                raise DataFileException(fail_msg)

            # Write length of block
            self.encoder.write_long(compressed_data_length)

            # Write block
            self.writer.write(compressed_data)
            
            # Write CRC32 checksum for Snappy
            if self.get_meta(CODEC_KEY) == 'snappy':
                self.encoder.write_crc32(uncompressed_data)

            # write sync marker
            self.writer.write(self.sync_marker)

            # reset buffer
            self.buffer_writer.truncate(0) 
            self.block_count = 0

    def append(self, datum):
        """Append a datum to the file."""
        self.datum_writer.write(datum, self.buffer_encoder)
        self.block_count += 1

        # if the data to write is larger than the sync interval, write the block
        if self.buffer_writer.tell() >= SYNC_INTERVAL:
            self._write_block()

    def sync(self):
        """
        Return the current position as a value that may be passed to
        DataFileReader.seek(long). Forces the end of the current block,
        emitting a synchronization marker.
        """
        self._write_block()
        return self.writer.tell()

    def flush(self):
        """Flush the current state of the file, including metadata."""
        self._write_block()
        self.writer.flush()

    def close(self):
        """Close the file."""
        self.flush()
        self.writer.close()


class DataFileReader(six.Iterator):
    """Read files written by DataFileWriter."""
    # TODO(hammer): allow user to specify expected schema?
    # TODO(hammer): allow user to specify the encoder
    def __init__(self, reader, datum_reader):
        self._reader = reader
        self._raw_decoder = io.BinaryDecoder(reader)
        self._datum_decoder = None # Maybe reset at every block.
        self._datum_reader = datum_reader
        
        # read the header: magic, meta, sync
        self._read_header()

        # ensure codec is valid
        self.codec = self.get_meta(CODEC_KEY)
        if self.codec is None:
            self.codec = "null"
        if self.codec not in VALID_CODECS:
            raise DataFileException('Unknown codec: %s.' % self.codec)

        # get file length
        self._file_length = self.determine_file_length()

        # get ready to read
        self._block_count = 0
        self.datum_reader.writers_schema = schema.parse(self.get_meta(SCHEMA_KEY))

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # Perform a close if there's no exception
        if type is None:
            self.close()

    def __iter__(self):
        return self

    # read-only properties
    reader = property(lambda self: self._reader)
    raw_decoder = property(lambda self: self._raw_decoder)
    datum_decoder = property(lambda self: self._datum_decoder)
    datum_reader = property(lambda self: self._datum_reader)
    sync_marker = property(lambda self: self._sync_marker)
    meta = property(lambda self: self._meta)
    file_length = property(lambda self: self._file_length)

    # read/write properties
    def set_block_count(self, new_val):
        self._block_count = new_val
    block_count = property(lambda self: self._block_count, set_block_count)

    # utility functions to read/write metadata entries
    def get_meta(self, key):
        return self._meta.get(key).decode('utf-8')
    def set_meta(self, key, val):
        self._meta[key] = val.encode('utf-8')

    def determine_file_length(self):
        """
        Get file length and leave file cursor where we found it.
        """
        remember_pos = self.reader.tell()
        self.reader.seek(0, 2)
        file_length = self.reader.tell()
        self.reader.seek(remember_pos)
        return file_length

    def is_EOF(self):
        return self.reader.tell() == self.file_length

    def _read_header(self):
        # seek to the beginning of the file to get magic block
        self.reader.seek(0, 0)

        # read header into a dict
        header = self.datum_reader.read_data(
            META_SCHEMA, META_SCHEMA, self.raw_decoder)

        # check magic number
        if header.get('magic') != MAGIC:
            fail_msg = "Not an Avro data file: %s doesn't match %s."\
                                 % (header.get('magic'), MAGIC)
            raise schema.AvroException(fail_msg)

        # set metadata
        self._meta = header['meta']

        # set sync marker
        self._sync_marker = header['sync']

    def _read_block_header(self):
        self.block_count = self.raw_decoder.read_long()
        if self.codec == "null":
            # Skip a long; we don't need to use the length.
            self.raw_decoder.skip_long()
            self._datum_decoder = self._raw_decoder
        elif self.codec == 'deflate':
            # Compressed data is stored as (length, data), which
            # corresponds to how the "bytes" type is encoded.
            data = self.raw_decoder.read_bytes()
            # -15 is the log of the window size; negative indicates
            # "raw" (no zlib headers) decompression.    See zlib.h.
            uncompressed = zlib.decompress(data, -15)
            self._datum_decoder = io.BinaryDecoder(StringIO(uncompressed))
        elif self.codec == 'snappy':
            # Compressed data includes a 4-byte CRC32 checksum
            length = self.raw_decoder.read_long()
            data = self.raw_decoder.read(length - 4)
            uncompressed = snappy.decompress(data)
            self._datum_decoder = io.BinaryDecoder(StringIO(uncompressed))
            self.raw_decoder.check_crc32(uncompressed);
        else:
            raise DataFileException("Unknown codec: %r" % self.codec)

    def _skip_sync(self):
        """
        Read the length of the sync marker; if it matches the sync marker,
        return True. Otherwise, seek back to where we started and return False.
        """
        proposed_sync_marker = self.reader.read(SYNC_SIZE)
        if proposed_sync_marker != self.sync_marker:
            self.reader.seek(-SYNC_SIZE, 1)
            return False
        else:
            return True

    # TODO(hammer): handle block of length zero
    # TODO(hammer): clean this up with recursion
    def __next__(self):
        """Return the next datum in the file."""
        if self.block_count == 0:
            if self.is_EOF():
                raise StopIteration
            elif self._skip_sync():
                if self.is_EOF(): raise StopIteration
                self._read_block_header()
            else:
                self._read_block_header()

        datum = self.datum_reader.read(self.datum_decoder) 
        self.block_count -= 1
        return datum

    def close(self):
        """Close this reader."""
        self.reader.close()

def generate_sixteen_random_bytes():
    try:
        import os
        return os.urandom(16)
    except:
        import random
        return [ chr(random.randrange(256)) for i in range(16) ]



# Copyright (C) 2017 Pluralsight LLC
'''Fast Cython extension for reading / writing and validating AVRO records.

The main edge this code has is that it parses the schema only once and creates
a reader/writer call tree from the schema shape. All reads and writes then
no longer consult the schema saving lookups.'''

import six
INT_MIN_VALUE = -(1 << 31)
INT_MAX_VALUE = (1 << 31) - 1
LONG_MIN_VALUE = -(1 << 63)
LONG_MAX_VALUE = (1 << 63) - 1

cdef long long read_long(fo):
    '''Read a long using zig-zag binary encoding'''
    cdef:
        unsigned long long accum
        int temp_datum
        char* c_raw
        int shift = 7
    raw = fo.read(1)
    c_raw = raw
    temp_datum = <int>c_raw[0]
    accum = temp_datum & 0x7F
    while (temp_datum & 0x80) != 0:
        raw = fo.read(1)
        c_raw = raw
        temp_datum = <int>c_raw[0]
        accum |= (temp_datum & 0x7F) << shift
        shift += 7
    return (accum >> 1) ^ -(accum & 1)

cdef bytes read_bytes(fo):
    '''Bytes are a marker for length of bytes and then binary data'''
    return fo.read(read_long(fo))


cdef read_null(fo):
    """
    null is written as zero bytes
    """
    return None


def read_boolean(fo):
    """
    a boolean is written as a single byte 
    whose value is either 0 (false) or 1 (true).
    """
    return fo.read(1) == b'\x01'

cdef float read_float(fo):
    """
    A float is written as 4 bytes.
    The float is converted into a 32-bit integer using a method equivalent to
    Java's floatToIntBits and then encoded in little-endian format.
    """
    data = fo.read(4)
    cdef char* y = data
    return (<float*>y)[0]

cdef double read_double(fo):
    """
    A double is written as 8 bytes.
    The double is converted into a 64-bit integer using a method equivalent to
    Java's doubleToLongBits and then encoded in little-endian format.
    """
    data = fo.read(8)
    cdef char* y = data
    return (<double*>y)[0]

cdef unicode read_utf8(fo):
    """
    A string is encoded as a long followed by
    that many bytes of UTF-8 encoded character data.
    """
    byte_data = read_bytes(fo)
    return unicode(byte_data, "utf-8")

# ======================================================================
from collections import namedtuple
ReadField = namedtuple('ReadField', ['name', 'reader', 'skip'])
WriteField = namedtuple('WriteField', ['name', 'writer'])

cpdef unicode get_type(schema):
    if isinstance(schema, list):
        return u"union"
    elif isinstance(schema, dict):
        return unicode(schema['type'])  # "record"
    else:
        return unicode(schema)


def make_union_reader(union_schema):
    cdef list readers = [get_reader(schema) for schema in union_schema]

    def union_reader(fo):
        '''Read the long index for which schema to process, then use that'''
        union_index = read_long(fo)
        return readers[union_index](fo)
    union_reader.__reduce__ = lambda: (make_union_reader, (union_schema,))
    return union_reader


def make_record_reader(schema):
    cdef list fields = [ReadField(field['name'], get_reader(field['type']), get_type(field['type']) == 'skip') for field in schema['fields']]

    def record_reader(fo):
        return {field.name: field.reader(fo) for field in fields if not (field.skip and field.reader(fo) is None)}
    record_reader.__reduce__ = lambda: (make_record_reader, (schema,))
    return record_reader


def make_enum_reader(schema):
    cdef list symbols = schema['symbols']

    def enum_reader(fo):
        return symbols[read_long(fo)]
    enum_reader.__reduce__ = lambda: (make_enum_reader, (schema,))
    return enum_reader

def make_array_reader(schema):
    item_reader = get_reader(schema['items'])
    def array_reader(fo):
        cdef long block_count
        cdef list read_items = []
        block_count = read_long(fo)
        while block_count != 0:
            if block_count < 0:
                block_count = -block_count
                block_size = read_long(fo)
            for i in range(block_count):
                read_items.append(item_reader(fo))
            block_count = read_long(fo)
        return read_items
    array_reader.__reduce__ = lambda: (make_array_reader, (schema,))
    return array_reader

def make_map_reader(schema):
    value_reader = get_reader(schema['values'])

    def map_reader(fo):
        cdef long block_count = read_long(fo)
        cdef dict read_items = {}
        while block_count != 0:
            if block_count < 0:
                block_count = -block_count
                block_size = read_long(fo)
            for _ in range(block_count):
                key = read_utf8(fo)
                read_items[key] = value_reader(fo)
            block_count = read_long(fo)
        return read_items
    map_reader.__reduce__ = lambda: (make_map_reader, (schema,))
    return map_reader

def make_fixed_reader(schema):
    cdef long size = schema['size']

    def fixed_reader(fo):
        return fo.read(size)
    fixed_reader.__reduce__ = lambda: (make_fixed_reader, (schema,))
    return fixed_reader

def make_null_reader(schema):
    return read_null

def make_string_reader(schema):
    return read_utf8

def make_boolean_reader(schema):
    return read_boolean

def make_double_reader(schema):
    return read_double

def make_long_reader(schema):
    return read_long

def make_byte_reader(schema):
    return read_bytes

def make_float_reader(schema):
    return read_float


def make_skip_reader(schema):
    # this will create a regular reader that will iterate the bytes
    # in the avro stream properly
    value_reader = get_reader(schema['value'])
    def read_skip(fo):
        value_reader(fo)
        return None
    read_skip.__reduce__ = lambda: (make_skip_reader, (schema,))
    return read_skip


def make_default_reader(schema):
    value = schema["value"]
    def read_default(fo):
        return value
    read_default.__reduce__ = lambda: (make_default_reader, (schema,))
    return read_default


reader_type_map = {
    'union': make_union_reader,
    'record': make_record_reader,
    'null': make_null_reader,
    'string': make_string_reader,
    'boolean': make_boolean_reader,
    'double': make_double_reader,
    'float': make_float_reader,
    'long': make_long_reader,
    'bytes': make_byte_reader,
    'int': make_long_reader,
    'fixed': make_fixed_reader,
    'enum': make_enum_reader,
    'array': make_array_reader,
    'map': make_map_reader,
    'skip': make_skip_reader,
    'default': make_default_reader
}

schema_cache = {}

class ReaderPlaceholder(object):
    def __init__(self):
        self.reader = None

    def __call__(self, fo):
        return self.reader(fo)

def get_reader(schema):
    cdef unicode schema_type = get_type(schema)
    if schema_type in ('record', 'fixed', 'enum'):
        placeholder = ReaderPlaceholder()
        # using a placeholder because this is recursive and the reader isn't defined
        # yet and nested records might refer to this parent schema name
        namespace = schema.get('namespace')
        record_name = schema.get('name')
        if namespace:
           namspace_record_name = '.'.join([namespace, record_name])
        else:
            namspace_record_name = record_name
        schema_cache[namspace_record_name] = placeholder
        reader = reader_type_map[schema_type](schema)
        # now that we've returned, assign the reader to the placeholder
        # so that the execution will work
        placeholder.reader = reader
        return reader
    try:
        reader = reader_type_map[schema_type](schema)
    except KeyError:
        reader = schema_cache[schema_type]

    return reader

# ======================================================================


cdef void write_int(outbuf, long long signed_datum):
    """int and long values are written using variable-length, zig-zag coding.
    """
    cdef:
        unsigned long long datum
        char temp_datum
    datum = (signed_datum << 1) ^ (signed_datum >> 63)
    while datum > 127:
        temp_datum = (datum & 0x7f) | 0x80
        outbuf.write((<char *>&temp_datum)[:sizeof(char)])
        datum >>= 7
    outbuf.write((<char *>&datum)[:sizeof(char)])

write_long = write_int


cdef void write_bytes(outbuf, datum):
    """
    Bytes are encoded as a long followed by that many bytes of data. 
    """
    cdef long byte_count = len(datum)
    write_long(outbuf, byte_count)
    outbuf.write(datum)


# except *
cdef void write_utf8(outbuf, datum):
    """
    Bytes are encoded as a long followed by that many bytes of data.
    """
    write_bytes(outbuf, datum.encode("utf-8"))


cdef void write_float(outbuf, float datum):
    """
    A float is written as 4 bytes.
    The float is converted into a 32-bit integer using a method equivalent to
    Java's floatToIntBits and then encoded in little-endian format.
    """
    outbuf.write((<char *>&datum)[:sizeof(float)])


cdef void write_double(outbuf, double datum):
    """
    A double is written as 8 bytes.
    The double is converted into a 64-bit integer using a method equivalent to
    Java's doubleToLongBits and then encoded in little-endian format.
    """
    outbuf.write((<char *>&datum)[:sizeof(double)])


cdef void write_null(outbuf, datum):
    pass


cdef void write_fixed(outbuf, datum):
    """A fixed writer writes out exactly the bytes up to a count"""
    outbuf.write(datum)


cdef write_boolean(outbuf, char datum):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true)."""
    cdef char x = 1 if datum else 0
    outbuf.write((<char *>&x)[:sizeof(char)])


avro_to_py = {
    u"string": unicode,
    u"int": int,
    u"long": int,
    u"boolean": bool,
    u"null": type(None),
    u"float": float,
    u"double": float,
    u"array": list,
    u"record": dict,
    u"enum": unicode,
    u"fixed": str,
    u"map": dict
}

py_to_avro = {
    unicode: u'string',
    str: u'string',
    int: u'int',
    long: u'long',
    bool: u'boolean',
    type(None): u'null',
    float: u'double',
    list: u'array',
    dict: u'record'
}

# ===============================
CheckField = namedtuple('CheckField', ['name', 'check'])

def get_check(schema):
    cdef unicode schema_type = get_type(schema)
    return check_type_map[schema_type](schema)


def make_record_check(schema):
    cdef list fields = [CheckField(field['name'], get_check(field['type'])) for field in schema['fields']]
    def record_check(datum):
        return isinstance(datum, dict) and all([field.check(datum.get(field.name)) for field in fields])
    return record_check


def make_enum_check(schema):
    cdef list symbols = schema['symbols']
    def enum_check(datum):
        return datum in symbols
    return enum_check


def make_null_check(schema):
    return lambda datum: datum is None

def check_string(datum):
    return isinstance(datum, basestring)

def make_string_check(schema):
    return check_string

def make_long_check(schema):
    return lambda datum: isinstance(datum, int) or isinstance(datum, long)

def make_boolean_check(schema):
    return lambda datum: isinstance(datum, bool)

# def make_boolean_check(schema):
#     return lambda datum: isinstance(datum, boolean)

def make_float_check(schema):
    return lambda datum: isinstance(datum, int) or isinstance(datum, long) or isinstance(datum, float)

def make_double_check(schema):
    return lambda datum: isinstance(datum, int) or isinstance(datum, long) or isinstance(datum, float)

def make_byte_check(schema):
    return lambda datum: isinstance(datum, str) or isinstance(datum, bytes)

def make_array_check(schema):
    item_check = get_check(schema['items'])
    def array_check(datum):
        return all([item_check(item) for item in datum])
    return array_check

def make_union_check(union_schema):
    cdef list union_checks = [get_check(schema) for schema in union_schema]
    def union_check(datum):
        return any([check(datum) for check in union_checks])
    return union_check

def make_fixed_check(schema):
    cdef int size = schema['size']
    def fixed_check(datum):
        return (isinstance(datum, str) or isinstance(datum, bytes)) and len(datum) == size
    return fixed_check

def make_map_check(schema):
    map_value_check = get_check(schema['values'])
    def map_check(datum):
        return isinstance(datum, dict) and all([check_string(key) and map_value_check(value) for key, value in datum.items()])
    return map_check

check_type_map = {
    'union': make_union_check,
    'record': make_record_check,
    'null': make_null_check,
    'string': make_string_check,
    'boolean': make_boolean_check,
    'double': make_double_check,
    'float': make_float_check,
    'long': make_long_check,
    'bytes': make_byte_check,
    'int': make_long_check,
    'fixed': make_fixed_check,
    'enum': make_enum_check,
    'array': make_array_check,
    'map': make_map_check
}

# ====================

def make_union_writer(union_schema):
    cdef list type_list = [get_type(schema) for schema in union_schema]
    # cdef dict writer_lookup
    # cdef list record_list
    cdef dict writer_lookup_dict
    cdef char simple_union
    cdef list lookup_result
    cdef long idx

    # if there's more than one kind of record in the union
    # or if there's a string, enum or fixed combined in the union
    # or there is both a record and a map in the union then it's no longer
    # simple. The reason is that reocrds and maps both correspond to python
    # dict so a simple python type lookup isn't enough to schema match.
    # enums, strings and fixed are all python data type unicode or string
    # so those won't work either when mixed
    simple_union = not(type_list.count('record') > 1 or
                      len(set(type_list) & set(['string', 'enum', 'fixed'])) > 1 or
                      len(set(type_list) & set(['record', 'map'])) > 1)

    if simple_union:
        writer_lookup_dict = {avro_to_py[get_type(schema)]: (idx, get_writer(schema)) for idx, schema in enumerate(union_schema)}
        if int in writer_lookup_dict:
            writer_lookup_dict[long] = writer_lookup_dict[int]
        if unicode in writer_lookup_dict:
            writer_lookup_dict[str] = writer_lookup_dict[unicode]
        # warning, this will fail if there's both a long and int in a union
        # or a float and a double in a union (which is valid but nonsensical
        # in python but valid in avro)
        def simple_writer_lookup(datum):
            return writer_lookup_dict[type(datum)]

        writer_lookup = simple_writer_lookup
    else:
        writer_lookup_dict = {}
        for idx, schema in enumerate(union_schema):
            python_type = avro_to_py[get_type(schema)]
            if python_type in writer_lookup_dict:
                writer_lookup_dict[python_type] = writer_lookup_dict[python_type] + [(idx, get_check(schema), get_writer(schema))]
            else:
                writer_lookup_dict[python_type] = [(idx, get_check(schema), get_writer(schema))]

        if int in writer_lookup_dict:
            writer_lookup_dict[long] = writer_lookup_dict[int]
        if unicode in writer_lookup_dict:
            writer_lookup_dict[str] = writer_lookup_dict[unicode]

        def complex_writer_lookup(datum):
            cdef:
                long idx
                list lookup_result
            lookup_result = writer_lookup_dict[type(datum)]
            if len(lookup_result) == 1:
                idx, get_check, writer = lookup_result[0]
            else:
                for idx, get_check, writer in lookup_result:
                    if get_check(datum):
                        break
                else:
                    raise TypeError("No matching schema for datum: {}".format(datum))
            return idx, writer

        writer_lookup = complex_writer_lookup

    def write_union(outbuf, datum):
        idx, data_writer = writer_lookup(datum)
        write_long(outbuf, idx)
        data_writer(outbuf, datum)
    write_union.__reduce__ = lambda: (make_union_writer, (union_schema,))
    return write_union

def make_enum_writer(schema):
    cdef list symbols = schema['symbols']

    # the datum can be str or unicode?
    def write_enum(outbuf, basestring datum):
        cdef int enum_index = symbols.index(datum)
        write_int(outbuf, enum_index)
    write_enum.__reduce__ = lambda: (make_enum_writer, (schema,))
    return write_enum


def make_record_writer(schema):
    cdef list fields = [WriteField(field['name'], get_writer(field['type'])) for field in schema['fields']]

    def write_record(outbuf, datum):
        for field in fields:
            field.writer(outbuf, datum.get(field.name))
    write_record.__reduce__ = lambda: (make_record_writer, (schema,))
    return write_record


def make_array_writer(schema):
    item_writer = get_writer(schema['items'])

    def write_array(outbuf, list datum):
        cdef long item_count = len(datum)
        if item_count > 0:
            write_long(outbuf, item_count)
        for item in datum:
            item_writer(outbuf, item)
        write_long(outbuf, 0)
    write_array.__reduce__ = lambda: (make_array_writer, (schema,))
    return write_array


def make_map_writer(schema):
    map_value_writer = get_writer(schema['values'])

    def write_map(outbuf, datum):
        cdef long item_count = len(datum)
        if item_count > 0:
            write_long(outbuf, item_count)
        for key, val in datum.iteritems():
            write_utf8(outbuf, key)
            map_value_writer(outbuf, val)
        write_long(outbuf, 0)
    write_map.__reduce__ = lambda: (make_map_writer, (schema,))
    return write_map


def make_boolean_writer(schema):
    '''Create a boolean writer, adds a validation step before the actual
    write function'''
    def checked_boolean_writer(outbuf, datum):
        if not isinstance(datum, bool):
            raise TypeError("Not a boolean value")
        write_boolean(outbuf, datum)
    return checked_boolean_writer


def make_fixed_writer(schema):
    '''A writer that must write X bytes defined by the schema'''
    cdef long size = schema['size']
    # note: not a char* because those are null terminated and fixed
    # has no such limitation
    def checked_write_fixed(outbuf, datum):
        if len(datum) != size:
            raise TypeError("Size Mismatch for Fixed data")
        write_fixed(outbuf, datum)
    return checked_write_fixed


def make_int_writer(schema):
    '''Create a int writer, adds a validation step before the actual
    write function to make sure the int value doesn't overflow'''
    def checked_int_write(outbuf, datum):
        if not (isinstance(datum, six.integer_types)
                        and INT_MIN_VALUE <= datum <= INT_MAX_VALUE):
            raise TypeError("Non integer value or overflow")
        write_long(outbuf, datum)
    return checked_int_write


def make_long_writer(schema):
    '''Create a long writer, adds a validation step before the actual
    write function to make sure the long value doesn't overflow'''
    def checked_long_write(outbuf, datum):
        if not (isinstance(datum, six.integer_types)
                        and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE):
            raise TypeError("Non integer value or overflow")
        write_long(outbuf, datum)
    return checked_long_write


def make_string_writer(schema):
    def checked_string_writer(outbuf, datum):
        if not isinstance(datum, six.string_types):
            raise TypeError("Non string value")
        write_utf8(outbuf, datum)
    return checked_string_writer


def make_byte_writer(schema):
    return write_bytes


def make_float_writer(schema):
    return write_float


def make_double_writer(schema):
    return write_double


def make_null_writer(schema):
    return write_null


# writer
writer_type_map = {
    'union': make_union_writer,
    'record': make_record_writer,
    'null': make_null_writer,
    'string': make_string_writer,
    'boolean': make_boolean_writer,
    'double': make_double_writer,
    'float': make_float_writer,
    'long': make_long_writer,
    'bytes': make_byte_writer,
    'int': make_int_writer,
    'fixed': make_fixed_writer,
    'enum': make_enum_writer,
    'array': make_array_writer,
    'map': make_map_writer
}


class WriterPlaceholder(object):
    def __init__(self):
        self.writer = None

    def __call__(self, fo, val):
        return self.writer(fo, val)


def get_writer(schema):
    cdef unicode schema_type = get_type(schema)
    if schema_type in ('record', 'fixed', 'enum'):
        placeholder = WriterPlaceholder()
        # using a placeholder because this is recursive and the reader isn't defined
        # yet and nested records might refer to this parent schema name
        namespace = schema.get('namespace')
        name = schema.get('name')
        if namespace:
           fullname = '.'.join([namespace, name])
        else:
            fullname = name
        schema_cache[fullname] = placeholder
        writer = writer_type_map[schema_type](schema)
        # now that we've returned, assign the reader to the placeholder
        # so that the execution will work
        placeholder.writer = writer
        return writer
    try:
        writer = writer_type_map[schema_type](schema)
    except KeyError:
        writer = schema_cache[schema_type]

    return writer





import struct
from binascii import crc32


class FastBinaryEncoder(object):
    """Write leaf values."""
    def __init__(self, writer):
        """
        writer is a Python object on which we can call write.
        """
        self.writer = writer

    def write(self, datum):
        self.writer.write(datum)

    def write_null(self, datum):
        pass

    def write_boolean(self, datum):
        write_boolean(self.writer, datum)

    def write_int(self, datum):
        write_int(self.writer, datum)

    def write_long(self, datum):
        write_long(self.writer, datum)

    def write_float(self, datum):
        write_float(self.writer, datum)

    def write_double(self, datum):
        write_double(self.writer, datum)

    def write_bytes(self, datum):
        write_bytes(self.writer, datum)

    def write_utf8(self, datum):
        write_utf8(self.writer, datum)

    def write_crc32(self, bytes):
        """
        A 4-byte, big-endian CRC32 checksum
        """
        self.writer.write(struct.pack("!I", crc32(bytes) & 0xffffffff))



class FastBinaryDecoder(object):
    """Read leaf values."""
    def __init__(self, reader):
        """
        reader is a Python object on which we can call read, seek, and tell.
        """
        self.reader = reader

    def read(self, n):
        return self.reader.read(n)

    def read_null(self):
        return None

    def read_boolean(self):
        return read_boolean(self.reader)

    def read_int(self):
        return read_long(self.reader)

    def read_long(self):
        return read_long(self.reader)

    def read_float(self):
        return read_float(self.reader)

    def read_double(self):
        return read_double(self.reader)

    def read_bytes(self):
        return read_bytes(self.reader)

    def read_utf8(self):
        return read_utf8(self.reader)

    def check_crc32(self, bytes):
        checksum = struct.unpack("!I", self.reader.read(4))[0]
        if crc32(bytes) & 0xffffffff != checksum:
            raise RuntimeError("Checksum failure")

    def skip_null(self):
        pass

    def skip_boolean(self):
        self.reader.read(1)

    def skip_int(self):
        read_long(self.reader)

    def skip_long(self):
        read_long(self.reader)

    def skip_float(self):
        read_float(self.reader)

    def skip_double(self):
        read_double(self.reader)

    def skip_bytes(self):
        read_bytes(self.reader)

    def skip_utf8(self):
        read_utf8(self.reader)

    def skip(self, n):
        self.reader.seek(self.reader.tell() + n)



# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import struct
from binascii import crc32
from spavro import schema
import six

# TODO(hammer): shouldn't ! be < for little-endian (according to spec?)
if sys.version_info >= (2, 5, 0):
    struct_class = struct.Struct
else:
    class SimpleStruct(object):
        def __init__(self, format):
            self.format = format

        def pack(self, *args):
            return struct.pack(self.format, *args)

        def unpack(self, *args):
            return struct.unpack(self.format, *args)
    struct_class = SimpleStruct

STRUCT_INT = struct_class('!I')         # big-endian unsigned int
STRUCT_LONG = struct_class('!Q')        # big-endian unsigned long long
STRUCT_FLOAT = struct_class('!f')     # big-endian float
STRUCT_DOUBLE = struct_class('!d')    # big-endian double
STRUCT_CRC32 = struct_class('>I')     # big-endian unsigned int


class BinaryDecoder(object):
    """Read leaf values."""
    def __init__(self, reader):
        """
        reader is a Python object on which we can call read, seek, and tell.
        """
        self._reader = reader

    # read-only properties
    reader = property(lambda self: self._reader)

    def read(self, n):
        """
        Read n bytes.
        """
        return self.reader.read(n)

    def read_null(self):
        """
        null is written as zero bytes
        """
        return None

    def read_boolean(self):
        """
        a boolean is written as a single byte 
        whose value is either 0 (false) or 1 (true).
        """
        return ord(self.read(1)) == 1

    def read_int(self):
        """
        int and long values are written using variable-length, zig-zag coding.
        """
        return self.read_long()

    def read_long(self):
        """
        int and long values are written using variable-length, zig-zag coding.
        """
        b = ord(self.read(1))
        n = b & 0x7F
        shift = 7
        while (b & 0x80) != 0:
            b = ord(self.read(1))
            n |= (b & 0x7F) << shift
            shift += 7
        datum = (n >> 1) ^ -(n & 1)
        return datum

    def read_float(self):
        """
        A float is written as 4 bytes.
        The float is converted into a 32-bit integer using a method equivalent to
        Java's floatToIntBits and then encoded in little-endian format.
        """
        bits = (((ord(self.read(1)) & 0xff)) |
            ((ord(self.read(1)) & 0xff) <<    8) |
            ((ord(self.read(1)) & 0xff) << 16) |
            ((ord(self.read(1)) & 0xff) << 24))
        return STRUCT_FLOAT.unpack(STRUCT_INT.pack(bits))[0]

    def read_double(self):
        """
        A double is written as 8 bytes.
        The double is converted into a 64-bit integer using a method equivalent to
        Java's doubleToLongBits and then encoded in little-endian format.
        """
        bits = (((ord(self.read(1)) & 0xff)) |
            ((ord(self.read(1)) & 0xff) <<    8) |
            ((ord(self.read(1)) & 0xff) << 16) |
            ((ord(self.read(1)) & 0xff) << 24) |
            ((ord(self.read(1)) & 0xff) << 32) |
            ((ord(self.read(1)) & 0xff) << 40) |
            ((ord(self.read(1)) & 0xff) << 48) |
            ((ord(self.read(1)) & 0xff) << 56))
        return STRUCT_DOUBLE.unpack(STRUCT_LONG.pack(bits))[0]

    def read_bytes(self):
        """
        Bytes are encoded as a long followed by that many bytes of data. 
        """
        return self.read(self.read_long())

    def read_utf8(self):
        """
        A string is encoded as a long followed by
        that many bytes of UTF-8 encoded character data.
        """
        return six.text_type(self.read_bytes(), "utf-8")

    def check_crc32(self, bytes):
        checksum = STRUCT_CRC32.unpack(self.read(4))[0];
        if crc32(bytes) & 0xffffffff != checksum:
            raise schema.AvroException("Checksum failure")

    def skip_null(self):
        pass

    def skip_boolean(self):
        self.skip(1)

    def skip_int(self):
        self.skip_long()

    def skip_long(self):
        b = ord(self.read(1))
        while (b & 0x80) != 0:
            b = ord(self.read(1))

    def skip_float(self):
        self.skip(4)

    def skip_double(self):
        self.skip(8)

    def skip_bytes(self):
        self.skip(self.read_long())

    def skip_utf8(self):
        self.skip_bytes()

    def skip(self, n):
        self.reader.seek(self.reader.tell() + n)


class BinaryEncoder(object):
    """Write leaf values."""
    def __init__(self, writer):
        """
        writer is a Python object on which we can call write.
        """
        self._writer = writer

    # read-only properties
    writer = property(lambda self: self._writer)

    def write(self, datum):
        """Write an abritrary datum."""
        self.writer.write(datum)

    def write_null(self, datum):
        """
        null is written as zero bytes
        """
        pass

    def write_boolean(self, datum):
        """
        a boolean is written as a single byte 
        whose value is either 0 (false) or 1 (true).
        """
        if datum:
            self.write(six.int2byte(1))
        else:
            self.write(six.int2byte(0))

    def write_int(self, datum):
        """
        int and long values are written using variable-length, zig-zag coding.        
        """
        self.write_long(datum);

    def write_long(self, datum):
        """
        int and long values are written using variable-length, zig-zag coding.
        """
        datum = (datum << 1) ^ (datum >> 63)
        while (datum & ~0x7F) != 0:
            self.write(six.int2byte((datum & 0x7f) | 0x80))
            datum >>= 7
        self.write(six.int2byte(datum))

    def write_float(self, datum):
        """
        A float is written as 4 bytes.
        The float is converted into a 32-bit integer using a method equivalent to
        Java's floatToIntBits and then encoded in little-endian format.
        """
        bits = STRUCT_INT.unpack(STRUCT_FLOAT.pack(datum))[0]
        self.write(six.int2byte((bits) & 0xFF))
        self.write(six.int2byte((bits >> 8) & 0xFF))
        self.write(six.int2byte((bits >> 16) & 0xFF))
        self.write(six.int2byte((bits >> 24) & 0xFF))

    def write_double(self, datum):
        """
        A double is written as 8 bytes.
        The double is converted into a 64-bit integer using a method equivalent to
        Java's doubleToLongBits and then encoded in little-endian format.
        """
        bits = STRUCT_LONG.unpack(STRUCT_DOUBLE.pack(datum))[0]
        self.write(six.int2byte((bits) & 0xFF))
        self.write(six.int2byte((bits >> 8) & 0xFF))
        self.write(six.int2byte((bits >> 16) & 0xFF))
        self.write(six.int2byte((bits >> 24) & 0xFF))
        self.write(six.int2byte((bits >> 32) & 0xFF))
        self.write(six.int2byte((bits >> 40) & 0xFF))
        self.write(six.int2byte((bits >> 48) & 0xFF))
        self.write(six.int2byte((bits >> 56) & 0xFF))

    def write_bytes(self, datum):
        """
        Bytes are encoded as a long followed by that many bytes of data. 
        """
        self.write_long(len(datum))
        self.write(struct.pack('%ds' % len(datum), datum))

    def write_utf8(self, datum):
        """
        A string is encoded as a long followed by
        that many bytes of UTF-8 encoded character data.
        """
        datum = datum.encode("utf-8")
        self.write_bytes(datum)

    def write_crc32(self, bytes):
        """
        A 4-byte, big-endian CRC32 checksum
        """
        self.write(STRUCT_CRC32.pack(crc32(bytes) & 0xffffffff))



# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Input/Output utilities, including:

 * i/o-specific constants
 * i/o-specific exceptions
 * schema validation
 * leaf value encoding and decoding
 * datum reader/writer stuff (?)

Also includes a generic representation for data, which
uses the following mapping:

    * Schema records are implemented as dict.
    * Schema arrays are implemented as list.
    * Schema maps are implemented as dict.
    * Schema strings are implemented as unicode.
    * Schema bytes are implemented as str.
    * Schema ints are implemented as int.
    * Schema longs are implemented as long.
    * Schema floats are implemented as float.
    * Schema doubles are implemented as float.
    * Schema booleans are implemented as bool.
"""
import struct
from spavro import schema
import sys
from binascii import crc32

try:
    import json
except ImportError:
    import simplejson as json

import six
# python3
try:
    long
except NameError:
    long = int

from spavro.schema_resolve import resolve
#
# Constants
#

INT_MIN_VALUE = -(1 << 31)
INT_MAX_VALUE = (1 << 31) - 1
LONG_MIN_VALUE = -(1 << 63)
LONG_MAX_VALUE = (1 << 63) - 1

# TODO(hammer): shouldn't ! be < for little-endian (according to spec?)
if sys.version_info >= (2, 5, 0):
    struct_class = struct.Struct
else:
    class SimpleStruct(object):
        def __init__(self, format):
            self.format = format

        def pack(self, *args):
            return struct.pack(self.format, *args)

        def unpack(self, *args):
            return struct.unpack(self.format, *args)
    struct_class = SimpleStruct

STRUCT_INT = struct_class('!I')         # big-endian unsigned int
STRUCT_LONG = struct_class('!Q')        # big-endian unsigned long long
STRUCT_FLOAT = struct_class('!f')     # big-endian float
STRUCT_DOUBLE = struct_class('!d')    # big-endian double
STRUCT_CRC32 = struct_class('>I')     # big-endian unsigned int

#
# Exceptions
#

from spavro.exceptions import AvroTypeException, SchemaResolutionException

#
# Validate
#


def validate(expected_schema, datum):
    """Determine if a python datum is an instance of a schema."""
    schema_type = expected_schema.type
    if schema_type == 'null':
        return datum is None
    elif schema_type == 'boolean':
        return isinstance(datum, bool)
    elif schema_type == 'string':
        return isinstance(datum, six.string_types)
    elif schema_type == 'bytes':
        return isinstance(datum, bytes)
    elif schema_type == 'int':
        return (isinstance(datum, six.integer_types)
                        and INT_MIN_VALUE <= datum <= INT_MAX_VALUE)
    elif schema_type == 'long':
        return (isinstance(datum, six.integer_types)
                        and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE)
    elif schema_type in ['float', 'double']:
        return (isinstance(datum, six.integer_types)
                        or isinstance(datum, float))
    elif schema_type == 'fixed':
        return isinstance(datum, bytes) and len(datum) == expected_schema.size
    elif schema_type == 'enum':
        return datum in expected_schema.symbols
    elif schema_type == 'array':
        return (isinstance(datum, list) and
            False not in [validate(expected_schema.items, d) for d in datum])
    elif schema_type == 'map':
        return (isinstance(datum, dict) and
            False not in [isinstance(k, six.string_types) for k in datum.keys()] and
            False not in
                [validate(expected_schema.values, v) for v in datum.values()])
    elif schema_type in ['union', 'error_union']:
        return True in [validate(s, datum) for s in expected_schema.schemas]
    elif schema_type in ['record', 'error', 'request']:
        return (isinstance(datum, dict) and
            False not in
                [validate(f.type, datum.get(f.name)) for f in expected_schema.fields])

#
# Decoder/Encoder
#
import logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)
use_fast = False
try:
    from spavro.fast_binary import get_reader, get_writer
    from spavro.fast_binary import FastBinaryEncoder, FastBinaryDecoder
    use_fast = True
except ImportError:
    log.warn("Failed to load spavro C extension, using default slow Encoder/Decoder")
from spavro.binary import BinaryEncoder as SlowBinaryEncoder, BinaryDecoder as SlowBinaryDecoder

#
# SlowDatumReader/Writer
#


class SlowDatumReader(object):
    """Deserialize Avro-encoded data into a Python data structure."""
    @staticmethod
    def check_props(schema_one, schema_two, prop_list):
        for prop in prop_list:
            if getattr(schema_one, prop) != getattr(schema_two, prop):
                return False
        return True

    @staticmethod
    def match_schemas(writers_schema, readers_schema):
        w_type = writers_schema.type
        r_type = readers_schema.type
        if 'union' in [w_type, r_type] or 'error_union' in [w_type, r_type]:
            return True
        elif (w_type in schema.PRIMITIVE_TYPES and r_type in schema.PRIMITIVE_TYPES
                    and w_type == r_type):
            return True
        elif (w_type == r_type == 'record' and
                    SlowDatumReader.check_props(writers_schema, readers_schema, 
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'error' and
                    SlowDatumReader.check_props(writers_schema, readers_schema, 
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'request'):
            return True
        elif (w_type == r_type == 'fixed' and 
                    SlowDatumReader.check_props(writers_schema, readers_schema, 
                                                                    ['fullname', 'size'])):
            return True
        elif (w_type == r_type == 'enum' and 
                    SlowDatumReader.check_props(writers_schema, readers_schema, 
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'map' and 
                    SlowDatumReader.check_props(writers_schema.values,
                                                                    readers_schema.values, ['type'])):
            return True
        elif (w_type == r_type == 'array' and 
                    SlowDatumReader.check_props(writers_schema.items,
                                                                    readers_schema.items, ['type'])):
            return True
        
        # Handle schema promotion
        if w_type == 'int' and r_type in ['long', 'float', 'double']:
            return True
        elif w_type == 'long' and r_type in ['float', 'double']:
            return True
        elif w_type == 'float' and r_type == 'double':
            return True
        return False

    def __init__(self, writers_schema=None, readers_schema=None):
        """
        As defined in the Avro specification, we call the schema encoded
        in the data the "writer's schema", and the schema expected by the
        reader the "reader's schema".
        """
        self._writers_schema = writers_schema
        self._readers_schema = readers_schema 

    # read/write properties
    def set_writers_schema(self, writers_schema):
        self._writers_schema = writers_schema
    writers_schema = property(lambda self: self._writers_schema,
                                                        set_writers_schema)
    def set_readers_schema(self, readers_schema):
        self._readers_schema = readers_schema
    readers_schema = property(lambda self: self._readers_schema,
                                                        set_readers_schema)
    
    def read(self, decoder):
        if self.readers_schema is None:
            self.readers_schema = self.writers_schema
        return self.read_data(self.writers_schema, self.readers_schema, decoder)

    def read_data(self, writers_schema, readers_schema, decoder):
        # schema matching
        if not SlowDatumReader.match_schemas(writers_schema, readers_schema):
            fail_msg = 'Schemas do not match.'
            raise SchemaResolutionException(fail_msg, writers_schema, readers_schema)

        # schema resolution: reader's schema is a union, writer's schema is not
        if (writers_schema.type not in ['union', 'error_union']
                and readers_schema.type in ['union', 'error_union']):
            for s in readers_schema.schemas:
                if SlowDatumReader.match_schemas(writers_schema, s):
                    return self.read_data(writers_schema, s, decoder)
            fail_msg = 'Schemas do not match.'
            raise SchemaResolutionException(fail_msg, writers_schema, readers_schema)

        # function dispatch for reading data based on type of writer's schema
        if writers_schema.type == 'null':
            return decoder.read_null()
        elif writers_schema.type == 'boolean':
            return decoder.read_boolean()
        elif writers_schema.type == 'string':
            return decoder.read_utf8()
        elif writers_schema.type == 'int':
            return decoder.read_int()
        elif writers_schema.type == 'long':
            return decoder.read_long()
        elif writers_schema.type == 'float':
            return decoder.read_float()
        elif writers_schema.type == 'double':
            return decoder.read_double()
        elif writers_schema.type == 'bytes':
            return decoder.read_bytes()
        elif writers_schema.type == 'fixed':
            return self.read_fixed(writers_schema, readers_schema, decoder)
        elif writers_schema.type == 'enum':
            return self.read_enum(writers_schema, readers_schema, decoder)
        elif writers_schema.type == 'array':
            return self.read_array(writers_schema, readers_schema, decoder)
        elif writers_schema.type == 'map':
            return self.read_map(writers_schema, readers_schema, decoder)
        elif writers_schema.type in ['union', 'error_union']:
            return self.read_union(writers_schema, readers_schema, decoder)
        elif writers_schema.type in ['record', 'error', 'request']:
            return self.read_record(writers_schema, readers_schema, decoder)
        else:
            fail_msg = "Cannot read unknown schema type: %s" % writers_schema.type
            raise schema.AvroException(fail_msg)

    def skip_data(self, writers_schema, decoder):
        if writers_schema.type == 'null':
            return decoder.skip_null()
        elif writers_schema.type == 'boolean':
            return decoder.skip_boolean()
        elif writers_schema.type == 'string':
            return decoder.skip_utf8()
        elif writers_schema.type == 'int':
            return decoder.skip_int()
        elif writers_schema.type == 'long':
            return decoder.skip_long()
        elif writers_schema.type == 'float':
            return decoder.skip_float()
        elif writers_schema.type == 'double':
            return decoder.skip_double()
        elif writers_schema.type == 'bytes':
            return decoder.skip_bytes()
        elif writers_schema.type == 'fixed':
            return self.skip_fixed(writers_schema, decoder)
        elif writers_schema.type == 'enum':
            return self.skip_enum(writers_schema, decoder)
        elif writers_schema.type == 'array':
            return self.skip_array(writers_schema, decoder)
        elif writers_schema.type == 'map':
            return self.skip_map(writers_schema, decoder)
        elif writers_schema.type in ['union', 'error_union']:
            return self.skip_union(writers_schema, decoder)
        elif writers_schema.type in ['record', 'error', 'request']:
            return self.skip_record(writers_schema, decoder)
        else:
            fail_msg = "Unknown schema type: %s" % writers_schema.type
            raise schema.AvroException(fail_msg)

    def read_fixed(self, writers_schema, readers_schema, decoder):
        """
        Fixed instances are encoded using the number of bytes declared
        in the schema.
        """
        return decoder.read(writers_schema.size)

    def skip_fixed(self, writers_schema, decoder):
        return decoder.skip(writers_schema.size)

    def read_enum(self, writers_schema, readers_schema, decoder):
        """
        An enum is encoded by a int, representing the zero-based position
        of the symbol in the schema.
        """
        # read data
        index_of_symbol = decoder.read_int()
        if index_of_symbol >= len(writers_schema.symbols):
            fail_msg = "Can't access enum index %d for enum with %d symbols"\
                                 % (index_of_symbol, len(writers_schema.symbols))
            raise SchemaResolutionException(fail_msg, writers_schema, readers_schema)
        read_symbol = writers_schema.symbols[index_of_symbol]

        # schema resolution
        if read_symbol not in readers_schema.symbols:
            fail_msg = "Symbol %s not present in Reader's Schema" % read_symbol
            raise SchemaResolutionException(fail_msg, writers_schema, readers_schema)

        return read_symbol

    def skip_enum(self, writers_schema, decoder):
        return decoder.skip_int()

    def read_array(self, writers_schema, readers_schema, decoder):
        """
        Arrays are encoded as a series of blocks.

        Each block consists of a long count value,
        followed by that many array items.
        A block with count zero indicates the end of the array.
        Each item is encoded per the array's item schema.

        If a block's count is negative,
        then the count is followed immediately by a long block size,
        indicating the number of bytes in the block.
        The actual count in this case
        is the absolute value of the count written.
        """
        read_items = []
        block_count = decoder.read_long()
        while block_count != 0:
            if block_count < 0:
                block_count = -block_count
                block_size = decoder.read_long()
            for i in range(block_count):
                read_items.append(self.read_data(writers_schema.items,
                                                                                 readers_schema.items, decoder))
            block_count = decoder.read_long()
        return read_items

    def skip_array(self, writers_schema, decoder):
        block_count = decoder.read_long()
        while block_count != 0:
            if block_count < 0:
                block_size = decoder.read_long()
                decoder.skip(block_size)
            else:
                for i in range(block_count):
                    self.skip_data(writers_schema.items, decoder)
            block_count = decoder.read_long()

    def read_map(self, writers_schema, readers_schema, decoder):
        """
        Maps are encoded as a series of blocks.

        Each block consists of a long count value,
        followed by that many key/value pairs.
        A block with count zero indicates the end of the map.
        Each item is encoded per the map's value schema.

        If a block's count is negative,
        then the count is followed immediately by a long block size,
        indicating the number of bytes in the block.
        The actual count in this case
        is the absolute value of the count written.
        """
        read_items = {}
        block_count = decoder.read_long()
        while block_count != 0:
            if block_count < 0:
                block_count = -block_count
                block_size = decoder.read_long()
            for i in range(block_count):
                key = decoder.read_utf8()
                read_items[key] = self.read_data(writers_schema.values,
                                                        readers_schema.values, decoder)
            block_count = decoder.read_long()
        return read_items

    def skip_map(self, writers_schema, decoder):
        block_count = decoder.read_long()
        while block_count != 0:
            if block_count < 0:
                block_size = decoder.read_long()
                decoder.skip(block_size)
            else:
                for i in range(block_count):
                    decoder.skip_utf8()
                    self.skip_data(writers_schema.values, decoder)
            block_count = decoder.read_long()

    def read_union(self, writers_schema, readers_schema, decoder):
        """
        A union is encoded by first writing a long value indicating
        the zero-based position within the union of the schema of its value.
        The value is then encoded per the indicated schema within the union.
        """
        # schema resolution
        index_of_schema = int(decoder.read_long())
        if index_of_schema >= len(writers_schema.schemas):
            fail_msg = "Can't access branch index %d for union with %d branches"\
                                 % (index_of_schema, len(writers_schema.schemas))
            raise SchemaResolutionException(fail_msg, writers_schema, readers_schema)
        selected_writers_schema = writers_schema.schemas[index_of_schema]
        
        # read data
        return self.read_data(selected_writers_schema, readers_schema, decoder)

    def skip_union(self, writers_schema, decoder):
        index_of_schema = int(decoder.read_long())
        if index_of_schema >= len(writers_schema.schemas):
            fail_msg = "Can't access branch index %d for union with %d branches"\
                                 % (index_of_schema, len(writers_schema.schemas))
            raise SchemaResolutionException(fail_msg, writers_schema)
        return self.skip_data(writers_schema.schemas[index_of_schema], decoder)

    def read_record(self, writers_schema, readers_schema, decoder):
        """
        A record is encoded by encoding the values of its fields
        in the order that they are declared. In other words, a record
        is encoded as just the concatenation of the encodings of its fields.
        Field values are encoded per their schema.

        Schema Resolution:
         * the ordering of fields may be different: fields are matched by name.
         * schemas for fields with the same name in both records are resolved
             recursively.
         * if the writer's record contains a field with a name not present in the
             reader's record, the writer's value for that field is ignored.
         * if the reader's record schema has a field that contains a default value,
             and writer's schema does not have a field with the same name, then the
             reader should use the default value from its field.
         * if the reader's record schema has a field with no default value, and 
             writer's schema does not have a field with the same name, then the
             field's value is unset.
        """
        # schema resolution
        readers_fields_dict = readers_schema.fields_dict
        read_record = {}
        for field in writers_schema.fields:
            readers_field = readers_fields_dict.get(field.name)
            if readers_field is not None:
                field_val = self.read_data(field.type, readers_field.type, decoder)
                read_record[field.name] = field_val
            else:
                self.skip_data(field.type, decoder)

        # fill in default values
        if len(readers_fields_dict) > len(read_record):
            writers_fields_dict = writers_schema.fields_dict
            for field_name, field in readers_fields_dict.items():
                if field_name not in writers_fields_dict:
                    if field.has_default:
                        field_val = self._read_default_value(field.type, field.default)
                        read_record[field.name] = field_val
                    else:
                        fail_msg = 'No default value for field %s' % field_name
                        raise SchemaResolutionException(fail_msg, writers_schema,
                                                                                        readers_schema)
        return read_record

    def skip_record(self, writers_schema, decoder):
        for field in writers_schema.fields:
            self.skip_data(field.type, decoder)

    def _read_default_value(self, field_schema, default_value):
        """
        Basically a JSON Decoder?
        """
        if field_schema.type == 'null':
            return None
        elif field_schema.type == 'boolean':
            return bool(default_value)
        elif field_schema.type == 'int':
            return int(default_value)
        elif field_schema.type == 'long':
            return long(default_value)
        elif field_schema.type in ['float', 'double']:
            return float(default_value)
        elif field_schema.type in ['enum', 'fixed', 'string', 'bytes']:
            return default_value
        elif field_schema.type == 'array':
            read_array = []
            for json_val in default_value:
                item_val = self._read_default_value(field_schema.items, json_val)
                read_array.append(item_val)
            return read_array
        elif field_schema.type == 'map':
            read_map = {}
            for key, json_val in default_value.items():
                map_val = self._read_default_value(field_schema.values, json_val)
                read_map[key] = map_val
            return read_map
        elif field_schema.type in ['union', 'error_union']:
            return self._read_default_value(field_schema.schemas[0], default_value)
        elif field_schema.type == 'record':
            read_record = {}
            for field in field_schema.fields:
                json_val = default_value.get(field.name)
                if json_val is None: json_val = field.default
                field_val = self._read_default_value(field.type, json_val)
                read_record[field.name] = field_val
            return read_record
        else:
            fail_msg = 'Unknown type: %s' % field_schema.type
            raise schema.AvroException(fail_msg)


class SlowDatumWriter(object):
    """SlowDatumWriter for generic python objects."""
    def __init__(self, writers_schema=None):
        self._writers_schema = writers_schema

    # read/write properties
    def set_writers_schema(self, writers_schema):
        self._writers_schema = writers_schema
    writers_schema = property(lambda self: self._writers_schema,
                                                        set_writers_schema)

    def write(self, datum, encoder):
        # validate datum
        if not validate(self.writers_schema, datum):
            raise AvroTypeException(self.writers_schema, datum)
        self.write_data(self.writers_schema, datum, encoder)

    def write_data(self, writers_schema, datum, encoder):
        # function dispatch to write datum
        if writers_schema.type == 'null':
            encoder.write_null(datum)
        elif writers_schema.type == 'boolean':
            encoder.write_boolean(datum)
        elif writers_schema.type == 'string':
            encoder.write_utf8(datum)
        elif writers_schema.type == 'int':
            encoder.write_int(datum)
        elif writers_schema.type == 'long':
            encoder.write_long(datum)
        elif writers_schema.type == 'float':
            encoder.write_float(datum)
        elif writers_schema.type == 'double':
            encoder.write_double(datum)
        elif writers_schema.type == 'bytes':
            encoder.write_bytes(datum)
        elif writers_schema.type == 'fixed':
            self.write_fixed(writers_schema, datum, encoder)
        elif writers_schema.type == 'enum':
            self.write_enum(writers_schema, datum, encoder)
        elif writers_schema.type == 'array':
            self.write_array(writers_schema, datum, encoder)
        elif writers_schema.type == 'map':
            self.write_map(writers_schema, datum, encoder)
        elif writers_schema.type in ['union', 'error_union']:
            self.write_union(writers_schema, datum, encoder)
        elif writers_schema.type in ['record', 'error', 'request']:
            self.write_record(writers_schema, datum, encoder)
        else:
            fail_msg = 'Unknown type: %s' % writers_schema.type
            raise schema.AvroException(fail_msg)

    def write_fixed(self, writers_schema, datum, encoder):
        """
        Fixed instances are encoded using the number of bytes declared
        in the schema.
        """
        encoder.write(datum)

    def write_enum(self, writers_schema, datum, encoder):
        """
        An enum is encoded by a int, representing the zero-based position
        of the symbol in the schema.
        """
        index_of_datum = writers_schema.symbols.index(datum)
        encoder.write_int(index_of_datum)

    def write_array(self, writers_schema, datum, encoder):
        """
        Arrays are encoded as a series of blocks.

        Each block consists of a long count value,
        followed by that many array items.
        A block with count zero indicates the end of the array.
        Each item is encoded per the array's item schema.

        If a block's count is negative,
        then the count is followed immediately by a long block size,
        indicating the number of bytes in the block.
        The actual count in this case
        is the absolute value of the count written.
        """
        if len(datum) > 0:
            encoder.write_long(len(datum))
            for item in datum:
                self.write_data(writers_schema.items, item, encoder)
        encoder.write_long(0)

    def write_map(self, writers_schema, datum, encoder):
        """
        Maps are encoded as a series of blocks.

        Each block consists of a long count value,
        followed by that many key/value pairs.
        A block with count zero indicates the end of the map.
        Each item is encoded per the map's value schema.

        If a block's count is negative,
        then the count is followed immediately by a long block size,
        indicating the number of bytes in the block.
        The actual count in this case
        is the absolute value of the count written.
        """
        if len(datum) > 0:
            encoder.write_long(len(datum))
            for key, val in datum.items():
                encoder.write_utf8(key)
                self.write_data(writers_schema.values, val, encoder)
        encoder.write_long(0)

    def write_union(self, writers_schema, datum, encoder):
        """
        A union is encoded by first writing a long value indicating
        the zero-based position within the union of the schema of its value.
        The value is then encoded per the indicated schema within the union.
        """
        # resolve union
        index_of_schema = -1
        for i, candidate_schema in enumerate(writers_schema.schemas):
            if validate(candidate_schema, datum):
                index_of_schema = i
        if index_of_schema < 0: raise AvroTypeException(writers_schema, datum)

        # write data
        encoder.write_long(index_of_schema)
        self.write_data(writers_schema.schemas[index_of_schema], datum, encoder)

    def write_record(self, writers_schema, datum, encoder):
        """
        A record is encoded by encoding the values of its fields
        in the order that they are declared. In other words, a record
        is encoded as just the concatenation of the encodings of its fields.
        Field values are encoded per their schema.
        """
        for field in writers_schema.fields:
            self.write_data(field.type, datum.get(field.name), encoder)

# ===================


class FastDatumReader(object):
    """Deserialize Avro-encoded data into a Python data structure, Fast Version.

    Uses the new spavro C extension but maintains the old API.
    """
    @staticmethod
    def check_props(schema_one, schema_two, prop_list):
        for prop in prop_list:
            if getattr(schema_one, prop) != getattr(schema_two, prop):
                return False
        return True

    @staticmethod
    def match_schemas(writers_schema, readers_schema):
        w_type = writers_schema.type
        r_type = readers_schema.type
        if 'union' in [w_type, r_type] or 'error_union' in [w_type, r_type]:
            return True
        elif (w_type in schema.PRIMITIVE_TYPES and r_type in schema.PRIMITIVE_TYPES
                    and w_type == r_type):
            return True
        elif (w_type == r_type == 'record' and
                    FastDatumReader.check_props(writers_schema, readers_schema,
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'error' and
                    FastDatumReader.check_props(writers_schema, readers_schema,
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'request'):
            return True
        elif (w_type == r_type == 'fixed' and
                    FastDatumReader.check_props(writers_schema, readers_schema,
                                                                    ['fullname', 'size'])):
            return True
        elif (w_type == r_type == 'enum' and
                    FastDatumReader.check_props(writers_schema, readers_schema,
                                                                    ['fullname'])):
            return True
        elif (w_type == r_type == 'map' and
                    FastDatumReader.check_props(writers_schema.values,
                                                                    readers_schema.values, ['type'])):
            return True
        elif (w_type == r_type == 'array' and
                    FastDatumReader.check_props(writers_schema.items,
                                                                    readers_schema.items, ['type'])):
            return True

        # Handle schema promotion
        if w_type == 'int' and r_type in ['long', 'float', 'double']:
            return True
        elif w_type == 'long' and r_type in ['float', 'double']:
            return True
        elif w_type == 'float' and r_type == 'double':
            return True
        return False

    def __init__(self, writers_schema=None, readers_schema=None):
        """
        As defined in the Avro specification, we call the schema encoded
        in the data the "writer's schema", and the schema expected by the
        reader the "reader's schema".
        """
        self.schema_cache = {}
        if readers_schema:
            self.readers_schema = readers_schema
        if writers_schema:
            self.writers_schema = writers_schema
        if writers_schema and not readers_schema:
            self.readers_schema = writers_schema

    @property
    def writers_schema(self):
        return self._writers_schema

    @writers_schema.setter
    def writers_schema(self, parsed_writer_schema):
        '''Take a wrtiers schema object from the old avro API and run validation
        and schema reolution (if applicable)

        Since the old API would set this after object construction, we hook
        here so we can run schema validation and resolution once on assignment.
        This also makes the reader function for the resolved schema and uses
        that for data reading.
        '''
        self._writers_schema = parsed_writer_schema

        # create the reader function from the schema
        if hasattr(self, 'readers_schema'):
            # to_json is a terrible method name for something
            # that returns a python dict! :/
            resolved_schema = resolve(parsed_writer_schema.to_json(), self.readers_schema.to_json())
        else:
            # if no reader schema, then the resolved schema is just the writers
            # schema
            resolved_schema = parsed_writer_schema.to_json()
            self.readers_schema = parsed_writer_schema
        self.read_datum = get_reader(resolved_schema)

        # schema matching
        if not FastDatumReader.match_schemas(self.writers_schema, self.readers_schema):
            fail_msg = 'Schemas do not match.'
            raise SchemaResolutionException(fail_msg, self.writers_schema, self.readers_schema)

    def read(self, decoder):
        # keeping the API the same, we wrap the read_datum in read
        # and pass the underlying file object to the read_datum function
        return self.read_datum(decoder.reader)

    def read_data(self, writers_schema, readers_schema, decoder):
        resolved_schema = resolve(writers_schema.to_json(), readers_schema.to_json())
        # try:
        #     datum_reader = self.schema_cache[str(schema)]
        # except KeyError:
        datum_reader = get_reader(resolved_schema)
        # self.schema_cache[str(schema)] = datum_reader
        return datum_reader(decoder.reader)


class FastDatumWriter(object):
    """FastDatumWriter for generic python objects."""
    def __init__(self, writers_schema=None):
        self.writers_schema = writers_schema
        self.schema_cache = {}

    @property
    def writers_schema(self):
        return self._writers_schema

    @writers_schema.setter
    def writers_schema(self, parsed_writer_schema):
        self._writers_schema = parsed_writer_schema
        if parsed_writer_schema:
            # to_json is a terrible method name for something
            # that returns a python dict! :/
            self.write_datum = get_writer(parsed_writer_schema.to_json())

    def write(self, datum, encoder):
        # validate datum
        try:
            self.write_datum(encoder.writer, datum)
        except TypeError as ex:
            log.error(self.write_datum)
            log.exception("type error")
            raise AvroTypeException(self.writers_schema, datum)

    def write_data(self, schema, datum, encoder):
        try:
            datum_writer = self.schema_cache[str(schema)]
        except KeyError:
            datum_writer = get_writer(schema.to_json())
            self.schema_cache[str(schema)] = datum_writer
        datum_writer(encoder.writer, datum)

if use_fast:
    DatumReader = FastDatumReader
    DatumWriter = FastDatumWriter
    BinaryEncoder = FastBinaryEncoder
    BinaryDecoder = FastBinaryDecoder
else:
    DatumReader = SlowDatumReader
    DatumWriter = SlowDatumWriter
    BinaryEncoder = SlowBinaryEncoder
    BinaryDecoder = SlowBinaryDecoder


# Modifications copyright (C) 2017 Pluralsight LLC
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the schema parsing logic.
"""
import unittest
from spavro import schema


def print_name(test_name):
    print('')
    print(test_name)
    print('=' * len(test_name))
    print('')


class ExampleSchema(object):
    def __init__(self, schema_string, valid, name='', comment=''):
        self._schema_string = schema_string
        self._valid = valid
        self._name = name or schema_string # default to schema_string for name
        self.comment = comment

    @property
    def schema_string(self):
        return self._schema_string

    @property
    def valid(self):
        return self._valid

    @property
    def name(self):
        return self._name

#
# Example Schemas
#


def make_primitive_examples():
    examples = []
    for type in schema.PRIMITIVE_TYPES:
        examples.append(ExampleSchema('"%s"' % type, True))
        examples.append(ExampleSchema('{"type": "%s"}' % type, True))
    return examples

PRIMITIVE_EXAMPLES = [
    ExampleSchema('"True"', False),
    ExampleSchema('True', False),
    ExampleSchema('{"no_type": "test"}', False),
    ExampleSchema('{"type": "panther"}', False),
] + make_primitive_examples()

FIXED_EXAMPLES = [
    ExampleSchema('{"type": "fixed", "name": "Test", "size": 1}', True),
    ExampleSchema("""\
        {"type": "fixed",
         "name": "MyFixed",
         "namespace": "org.apache.hadoop.avro",
         "size": 1}
        """, True),
    ExampleSchema("""\
        {"type": "fixed",
         "name": "Missing size"}
        """, False),
    ExampleSchema("""\
        {"type": "fixed",
         "size": 314}
        """, False),
]

ENUM_EXAMPLES = [
    ExampleSchema('{"type": "enum", "name": "Test", "symbols": ["A", "B"]}', True),
    ExampleSchema("""\
        {"type": "enum",
         "name": "Status",
         "symbols": "Normal Caution Critical"}
        """, False),
    ExampleSchema("""\
        {"type": "enum",
         "name": [ 0, 1, 1, 2, 3, 5, 8 ],
         "symbols": ["Golden", "Mean"]}
        """, False),
    ExampleSchema("""\
        {"type": "enum",
         "symbols" : ["I", "will", "fail", "no", "name"]}
        """, False),
    ExampleSchema("""\
        {"type": "enum",
         "name": "Test"
         "symbols" : ["AA", "AA"]}
        """, False),
]

ARRAY_EXAMPLES = [
    ExampleSchema('{"type": "array", "items": "long"}', True),
    ExampleSchema("""\
        {"type": "array",
         "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]}}
        """, True),
]

MAP_EXAMPLES = [
    ExampleSchema('{"type": "map", "values": "long"}', True),
    ExampleSchema("""\
        {"type": "map",
         "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]}}
        """, True),
]

UNION_EXAMPLES = [
    ExampleSchema('["string", "null", "long"]', True),
    ExampleSchema('["null", "null"]', False),
    ExampleSchema('["long", "long"]', False),
    ExampleSchema("""\
        [{"type": "array", "items": "long"}
         {"type": "array", "items": "string"}]
        """, False),
]

RECORD_EXAMPLES = [
    ExampleSchema("""\
        {"type": "record",
         "name": "Test",
         "fields": [{"name": "f",
                                 "type": "long"}]}
        """, True),
    ExampleSchema("""\
        {"type": "error",
         "name": "Test",
         "fields": [{"name": "f",
                                 "type": "long"}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "Node",
         "fields": [{"name": "label", "type": "string"},
                                {"name": "children",
                                 "type": {"type": "array", "items": "Node"}}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "Lisp",
         "fields": [{"name": "value",
                                 "type": ["null", "string",
                                                    {"type": "record",
                                                     "name": "Cons",
                                                     "fields": [{"name": "car", "type": "Lisp"},
                                                                            {"name": "cdr", "type": "Lisp"}]}]}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "HandshakeRequest",
         "namespace": "org.apache.avro.ipc",
         "fields": [{"name": "clientHash",
                                 "type": {"type": "fixed", "name": "MD5", "size": 16}},
                                {"name": "clientProtocol", "type": ["null", "string"]},
                                {"name": "serverHash", "type": "MD5"},
                                {"name": "meta", 
                                 "type": ["null", {"type": "map", "values": "bytes"}]}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "HandshakeResponse",
         "namespace": "org.apache.avro.ipc",
         "fields": [{"name": "match",
                                 "type": {"type": "enum",
                                                    "name": "HandshakeMatch",
                                                    "symbols": ["BOTH", "CLIENT", "NONE"]}},
                                {"name": "serverProtocol", "type": ["null", "string"]},
                                {"name": "serverHash",
                                 "type": ["null",
                                                    {"name": "MD5", "size": 16, "type": "fixed"}]},
                                {"name": "meta",
                                 "type": ["null", {"type": "map", "values": "bytes"}]}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "Interop",
         "namespace": "org.apache.avro",
         "fields": [{"name": "intField", "type": "int"},
                                {"name": "longField", "type": "long"},
                                {"name": "stringField", "type": "string"},
                                {"name": "boolField", "type": "boolean"},
                                {"name": "floatField", "type": "float"},
                                {"name": "doubleField", "type": "double"},
                                {"name": "bytesField", "type": "bytes"},
                                {"name": "nullField", "type": "null"},
                                {"name": "arrayField",
                                 "type": {"type": "array", "items": "double"}},
                                {"name": "mapField",
                                 "type": {"type": "map",
                                                    "values": {"name": "Foo",
                                                                         "type": "record",
                                                                         "fields": [{"name": "label",
                                                                                                 "type": "string"}]}}},
                                {"name": "unionField",
                                 "type": ["boolean",
                                                    "double",
                                                    {"type": "array", "items": "bytes"}]},
                                {"name": "enumField",
                                 "type": {"type": "enum",
                                                    "name": "Kind",
                                                    "symbols": ["A", "B", "C"]}},
                                {"name": "fixedField",
                                 "type": {"type": "fixed", "name": "MD5", "size": 16}},
                                {"name": "recordField",
                                 "type": {"type": "record",
                                                    "name": "Node",
                                                    "fields": [{"name": "label", "type": "string"},
                                                                         {"name": "children",
                                                                            "type": {"type": "array",
                                                                                             "items": "Node"}}]}}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "ipAddr",
         "fields": [{"name": "addr",
                                 "type": [{"name": "IPv6", "type": "fixed", "size": 16},
                                                    {"name": "IPv4", "type": "fixed", "size": 4}]}]}
        """, True),
    ExampleSchema("""\
        {"type": "record",
         "name": "Address",
         "fields": [{"type": "string"},
                                {"type": "string", "name": "City"}]}
        """, False),
    ExampleSchema("""\
        {"type": "record",
         "name": "Event",
         "fields": [{"name": "Sponsor"},
                                {"name": "City", "type": "string"}]}
        """, False),
    ExampleSchema("""\
        {"type": "record",
         "fields": "His vision, from the constantly passing bars,"
         "name", "Rainer"}
        """, False),
    ExampleSchema("""\
        {"name": ["Tom", "Jerry"],
         "type": "record",
         "fields": [{"name": "name", "type": "string"}]}
        """, False),
]

DOC_EXAMPLES = [
    ExampleSchema("""\
        {"type": "record",
         "name": "TestDoc",
         "doc":  "Doc string",
         "fields": [{"name": "name", "type": "string", 
                                 "doc" : "Doc String"}]}
        """, True),
    ExampleSchema("""\
        {"type": "enum", "name": "Test", "symbols": ["A", "B"],
         "doc": "Doc String"}
        """, True),
]

OTHER_PROP_EXAMPLES = [
    ExampleSchema("""\
        {"type": "record",
         "name": "TestRecord",
         "cp_string": "string",
         "cp_int": 1,
         "cp_array": [ 1, 2, 3, 4],
         "fields": [ {"name": "f1", "type": "string", "cp_object": {"a":1,"b":2} },
                                 {"name": "f2", "type": "long", "cp_null": null} ]}
        """, True),
    ExampleSchema("""\
         {"type": "map", "values": "long", "cp_boolean": true}
        """, True),
    ExampleSchema("""\
        {"type": "enum",
         "name": "TestEnum",
         "symbols": [ "one", "two", "three" ],
         "cp_float" : 1.0 }
        """,True),
    ExampleSchema("""\
        {"type": "long",
         "date": "true"}
        """, True)
]

EXAMPLES = PRIMITIVE_EXAMPLES
EXAMPLES += FIXED_EXAMPLES
EXAMPLES += ENUM_EXAMPLES
EXAMPLES += ARRAY_EXAMPLES
EXAMPLES += MAP_EXAMPLES
EXAMPLES += UNION_EXAMPLES
EXAMPLES += RECORD_EXAMPLES
EXAMPLES += DOC_EXAMPLES

VALID_EXAMPLES = [e for e in EXAMPLES if e.valid]

# TODO(hammer): refactor into harness for examples
# TODO(hammer): pretty-print(detailed output)
# TODO(hammer): make verbose flag
# TODO(hammer): show strack trace to user
# TODO(hammer): use logging module?


class TestSchema(unittest.TestCase):

    def test_correct_recursive_extraction(self):
        s = schema.parse('{"type": "record", "name": "X", "fields": [{"name": "y", "type": {"type": "record", "name": "Y", "fields": [{"name": "Z", "type": "X"}]}}]}')
        t = schema.parse(str(s.fields[0].type))
        # If we've made it this far, the subschema was reasonably stringified; it ccould be reparsed.
        self.assertEqual("X", t.fields[0].type.name)

    def test_parse(self):
        correct = 0
        for example in EXAMPLES:
            try:
                schema.parse(example.schema_string)
                if example.valid:
                    correct += 1
                else:
                    self.fail("Invalid schema was parsed: " + example.schema_string)
            except:
                if not example.valid: 
                    correct += 1
                else:
                    self.fail("Valid schema failed to parse: " + example.schema_string)

        fail_msg = "Parse behavior correct on %d out of %d schemas." % \
            (correct, len(EXAMPLES))
        self.assertEqual(correct, len(EXAMPLES), fail_msg)

    def test_valid_cast_to_string_after_parse(self):
        # Test that the string generated by an Avro Schema object
        # is, in fact, a valid Avro schema.
        print_name('TEST CAST TO STRING AFTER PARSE')
        correct = 0
        for example in VALID_EXAMPLES:
            schema_data = schema.parse(example.schema_string)
            schema.parse(str(schema_data))
            correct += 1

        fail_msg = "Cast to string success on %d out of %d schemas" % \
            (correct, len(VALID_EXAMPLES))
        self.assertEqual(correct, len(VALID_EXAMPLES), fail_msg)

    def test_equivalence_after_round_trip(self):
        # 1. Given a string, parse it to get Avro schema "original".
        # 2. Serialize "original" to a string and parse that string
        #      to generate Avro schema "round trip".
        # 3. Ensure "original" and "round trip" schemas are equivalent.
        print_name('TEST ROUND TRIP')
        correct = 0
        for example in VALID_EXAMPLES:
            original_schema = schema.parse(example.schema_string)
            round_trip_schema = schema.parse(str(original_schema))
            if original_schema == round_trip_schema:
                correct += 1
                debug_msg = "%s: ROUND TRIP SUCCESS" % example.name
            else:
                debug_msg = "%s: ROUND TRIP FAILURE" % example.name
                self.fail("Round trip failure: %s, %s, %s" % (example.name, original_schema, str(original_schema)))

        fail_msg = "Round trip success on %d out of %d schemas" % \
            (correct, len(VALID_EXAMPLES))
        self.assertEqual(correct, len(VALID_EXAMPLES), fail_msg)

    # TODO(hammer): more tests
    def test_fullname(self):
        # The fullname is determined in one of the following ways:
        #  * A name and namespace are both specified.  For example,
        #    one might use "name": "X", "namespace": "org.foo"
        #    to indicate the fullname "org.foo.X".
        #  * A fullname is specified.  If the name specified contains
        #    a dot, then it is assumed to be a fullname, and any
        #    namespace also specified is ignored.  For example,
        #    use "name": "org.foo.X" to indicate the
        #    fullname "org.foo.X".
        #  * A name only is specified, i.e., a name that contains no
        #    dots.  In this case the namespace is taken from the most
        #    tightly encosing schema or protocol.  For example,
        #    if "name": "X" is specified, and this occurs
        #    within a field of the record definition
        #    of "org.foo.Y", then the fullname is "org.foo.X".

        # References to previously defined names are as in the latter
        # two cases above: if they contain a dot they are a fullname, if
        # they do not contain a dot, the namespace is the namespace of
        # the enclosing definition.

        # Primitive type names have no namespace and their names may
        # not be defined in any namespace.  A schema may only contain
        # multiple definitions of a fullname if the definitions are
        # equivalent.
        print_name('TEST FULLNAME')

        # name and namespace specified
        fullname = schema.Name('a', 'o.a.h', None).fullname
        self.assertEqual(fullname, 'o.a.h.a')

        # fullname and namespace specified
        fullname = schema.Name('a.b.c.d', 'o.a.h', None).fullname
        self.assertEqual(fullname, 'a.b.c.d')

        # name and default namespace specified
        fullname = schema.Name('a', None, 'b.c.d').fullname
        self.assertEqual(fullname, 'b.c.d.a')

        # fullname and default namespace specified
        fullname = schema.Name('a.b.c.d', None, 'o.a.h').fullname
        self.assertEqual(fullname, 'a.b.c.d')

        # fullname, namespace, default namespace specified
        fullname = schema.Name('a.b.c.d', 'o.a.a', 'o.a.h').fullname
        self.assertEqual(fullname, 'a.b.c.d')

        # name, namespace, default namespace specified
        fullname = schema.Name('a', 'o.a.a', 'o.a.h').fullname
        self.assertEqual(fullname, 'o.a.a.a')

    def test_doc_attributes(self):
        print_name('TEST DOC ATTRIBUTES')
        correct = 0
        for example in DOC_EXAMPLES:
            original_schema = schema.parse(example.schema_string)
            if original_schema.doc is not None:
                correct += 1
            if original_schema.type == 'record':
                for f in original_schema.fields:
                    if f.doc is None:
                        self.fail("Failed to preserve 'doc' in fields: " + example.schema_string)
        self.assertEqual(correct,len(DOC_EXAMPLES))

    def test_other_attributes(self):
        print_name('TEST OTHER ATTRIBUTES')
        correct = 0
        props = {}
        for example in OTHER_PROP_EXAMPLES:
            original_schema = schema.parse(example.schema_string)
            round_trip_schema = schema.parse(str(original_schema))
            self.assertEqual(original_schema.other_props,round_trip_schema.other_props)
            if original_schema.type == "record":
                field_props = 0
                for f in original_schema.fields:
                    if f.other_props:
                        props.update(f.other_props)
                        field_props += 1
                self.assertEqual(field_props,len(original_schema.fields))
            if original_schema.other_props:
                props.update(original_schema.other_props)
                correct += 1
        for k in props:
            v = props[k]
            if k == "cp_boolean":
                self.assertEqual(type(v), bool)
            elif k == "cp_int":
                self.assertEqual(type(v), int)
            elif k == "cp_object":
                self.assertEqual(type(v), dict)
            elif k == "cp_float":
                self.assertEqual(type(v), float)
            elif k == "cp_array":
                self.assertEqual(type(v), list)
        self.assertEqual(correct,len(OTHER_PROP_EXAMPLES))

    def test_exception_is_not_swallowed_on_parse_error(self):
        print_name('TEST EXCEPTION NOT SWALLOWED ON PARSE ERROR')
        with self.assertRaises(schema.SchemaParseException) as context:
                schema.parse('/not/a/real/file')

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2017 Pluralsight LLC

# resolve schemas
from spavro.fast_binary import get_type
from spavro.exceptions import SchemaResolutionException


def get_field_by_name(fields, name):
    '''Take a list of avro fields, scan the fields using the field name and
    return that field.'''
    field_names = [field['name'] for field in fields]
    return fields[field_names.index(name)]


def resolve_record(writer, reader):
    '''Take a writer and reader schema and return a 'meta' schema that allows
    transforming a previously written record into a new read structure.'''
    fields = []
    if writer['name'] != reader['name']:
        raise SchemaResolutionException("Schemas not compatible record names don't match")
    record_name = reader['name']
    optional = {}
    if "namespace" in writer and "namespace" in reader:
        optional["namespace"] = reader["namespace"]

    writer_fields = [field['name'] for field in writer['fields']]
    reader_fields = [field['name'] for field in reader['fields']]
    # check for defaults for records that are in reader
    # but not in writer and vice versa
    reader_but_not_writer = (set(reader_fields) - set(writer_fields))
    writer_but_not_reader = (set(writer_fields) - set(reader_fields))
    both_reader_and_writer = (set(writer_fields) & set(reader_fields))
    # run through the fields in writer order
    for field in writer['fields']:
        if field['name'] in both_reader_and_writer:
            fields.append({"name": field['name'], "type": resolve(field['type'],
                get_field_by_name(reader['fields'], field['name'])['type'])})
        elif field['name'] in writer_but_not_reader:
            ### special skip type record
            fields.append({"name": field['name'], "type": {"type": "skip", "value": field['type']}})

    for field in reader['fields']:
        if field['name'] in reader_but_not_writer:
            try:
                fields.append({"name": field['name'], "type": {"type": "default", "value": field['default']}})
            except KeyError:
                raise SchemaResolutionException("Schemas not compatible, no default value for field in reader's record that's not present in writer's record")
    schema = {"type": "record", "fields": fields, "name": record_name}
    schema.update(optional)
    return schema


primitive_types = ('null', 'boolean', 'int', 'long', 'float', 'double', 'bytes', 'string' )


def resolve_array(writer, reader):
    '''Resolve a writer and reader array schema and recursively resolve the
    type for each array item schema'''
    return {'type': 'array', 'items': resolve(writer['items'], reader['items'])}


def resolve_map(writer, reader):
    '''Resolve a writer and reader map schema and resolve the type for the
    map's value schema'''
    return {'type': 'map', 'values': resolve(writer['values'], reader['values'])}


def resolve_enum(writer, reader):
    '''Compare a writer and reader enum and return a compatible enum'''
    if writer['name'] != reader['name']:
        raise SchemaResolutionException("Schemas not compatible, enum names don't match")
    if set(writer['symbols']) - set(reader['symbols']):
        raise SchemaResolutionException("Schemas not compatible, symbol in writer's enum not present in reader's enum")
    return {'type': 'enum', 'name': reader['name'], 'symbols': [symbol for symbol in writer['symbols']]}


def resolve_fixed(writer, reader):
    '''Take a fixed writer and reader schema and return the writers size value.
    '''
    if writer['name'] != reader['name'] or writer['size'] != reader['size']:
        raise SchemaResolutionException("Schemas not compatible, fixed names or sizes don't match")
    return {key: value for key, value in writer.items()}


def resolve_union(writer, reader):
    '''Take a writer union and a reader union, compare their types and return
    a read/write compatible union.

    A compatible read/write union has all of the writer's union schemas in the
    reader's schema.
    '''
    union = []
    for w_type in writer:
        for r_type in reader:
            try:
                merged = resolve(w_type, r_type)
                union.append(merged)
                break
            except SchemaResolutionException:
                # keep trying until we iterate through all read types
                continue
        else:
            # none of the read types matched the write type, this is an error
            raise SchemaResolutionException("Schema in writer's union not present in reader's union.")
    return union


promotable = ['int', 'long', 'float', 'double']


def resolve(writer, reader):
    '''Take a writer and a reader schema and return a meta schema that
    translates the writer's schema to the reader's schema.

    This handles skipping missing fields and default fills by creating
    non-standard 'types' for reader creation. These non-standard types are
    never surfaced out since they're not standard avro types but just used
    as an implementation detail for generating a write-compantible reader.'''
    writer_type = get_type(writer)
    reader_type = get_type(reader)

    if writer_type == reader_type:
        if reader_type in primitive_types:
            return reader
        if reader_type == 'array':
            return resolve_array(writer, reader)
        if reader_type == 'map':
            return resolve_map(writer, reader)
        if reader_type == 'enum':
            return resolve_enum(writer, reader)
        if reader_type == 'union':
            return resolve_union(writer, reader)
        if reader_type == "record":
            return resolve_record(writer, reader)
        if reader_type == "fixed":
            return resolve_fixed(writer, reader)
    else:
        # see if we've 'upgraded' to a union
        if reader_type == 'union':
            # if the writer type is in the reader's union
            # then just return the writer's schema
            if writer_type in [get_type(r) for r in reader]:
                type_index = [get_type(r) for r in reader].index(writer_type)
                return resolve(writer, reader[type_index])
            else:
                raise SchemaResolutionException("Writer schema not present in reader union")
        if writer_type in promotable and reader_type in promotable and promotable.index(writer_type) < promotable.index(reader_type):
            return writer
        raise SchemaResolutionException("Reader and Writer schemas are incompatible")



# Modifications copyright (C) 2017 Pluralsight LLC
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest

from spavro import schema
from spavro import io
from spavro import datafile

try:
    type(unicode)
except NameError:
    unicode = str

SCHEMAS_TO_VALIDATE = (
    ('"null"', None),
    ('"boolean"', True),
    ('"string"', unicode('adsfasdf09809dsf-=adsf')),
    ('"bytes"', b'12345abcd'),
    ('"int"', 1234),
    ('"long"', 1234),
    ('"float"', 1234.0),
    ('"double"', 1234.0),
    ('{"type": "fixed", "name": "Test", "size": 1}', b'B'),
    ('{"type": "enum", "name": "Test", "symbols": ["A", "B"]}', u'B'),
    ('{"type": "array", "items": "long"}', [1, 3, 2]),
    ('{"type": "map", "values": "long"}', {'a': 1, 'b': 3, 'c': 2}),
    ('["string", "null", "long"]', None),
    ("""\
     {"type": "record",
        "name": "Test",
        "fields": [{"name": "f", "type": "long"}]}
     """, {'f': 5}),
    ("""\
     {"type": "record",
        "name": "Lisp",
        "fields": [{"name": "value",
                                "type": ["null", "string",
                                                 {"type": "record",
                                                    "name": "Cons",
                                                    "fields": [{"name": "car", "type": "Lisp"},
                                                                         {"name": "cdr", "type": "Lisp"}]}]}]}
     """, {'value': {'car': {'value': 'head'}, 'cdr': {'value': None}}}),
)

FILENAME = 'test_datafile.out'
CODECS_TO_VALIDATE = ('null', 'deflate')
try:
    import snappy
    CODECS_TO_VALIDATE += ('snappy',)
except ImportError:
    print('Snappy not present, will skip testing it.')

# TODO(hammer): clean up written files with ant, not os.remove
class TestDataFile(unittest.TestCase):
    def test_round_trip(self):
        print('')
        print('TEST ROUND TRIP')
        print('===============')
        print('')
        correct = 0
        print(SCHEMAS_TO_VALIDATE)
        for i, (example_schema, datum) in enumerate(SCHEMAS_TO_VALIDATE):
            for codec in CODECS_TO_VALIDATE:
                print('')
                print('SCHEMA NUMBER %d' % (i + 1))
                print('================')
                print('')
                print('Schema: %s' % example_schema)
                print('Datum: %s' % datum)
                print('Codec: %s' % codec)

                # write data in binary to file 10 times
                writer = open(FILENAME, 'wb')
                datum_writer = io.DatumWriter()
                schema_object = schema.parse(example_schema)
                dfw = datafile.DataFileWriter(writer, datum_writer, schema_object, codec=codec)
                for datum_counter in range(10):
                    dfw.append(datum)
                dfw.close()

                # read data in binary from file
                reader = open(FILENAME, 'rb')
                datum_reader = io.DatumReader()
                dfr = datafile.DataFileReader(reader, datum_reader)
                round_trip_data = []
                for read_datum in dfr:
                    round_trip_data.append(read_datum)

                print('Round Trip Data: %s' % round_trip_data)
                print('Round Trip Data Length: %d' % len(round_trip_data))
                is_correct = [datum] * 10 == round_trip_data
                if is_correct:
                    correct += 1
                print('Correct Round Trip: %s' % is_correct)
                print('')
        os.remove(FILENAME)
        self.assertEquals(correct, len(CODECS_TO_VALIDATE) * len(SCHEMAS_TO_VALIDATE))

    def test_append(self):
        print('')
        print('TEST APPEND')
        print('===========')
        print('')
        correct = 0
        for i, (example_schema, datum) in enumerate(SCHEMAS_TO_VALIDATE):
            for codec in CODECS_TO_VALIDATE:
                print('')
                print('SCHEMA NUMBER %d' % (i + 1))
                print('================')
                print('')
                print('Schema: %s' % example_schema)
                print('Datum: %s' % datum)
                print('Codec: %s' % codec)

                # write data in binary to file once
                writer = open(FILENAME, 'wb')
                datum_writer = io.DatumWriter()
                schema_object = schema.parse(example_schema)
                dfw = datafile.DataFileWriter(writer, datum_writer, schema_object, codec=codec)
                dfw.append(datum)
                dfw.close()

                # open file, write, and close nine times
                for i in range(9):
                    writer = open(FILENAME, 'ab+')
                    dfw = datafile.DataFileWriter(writer, io.DatumWriter())
                    dfw.append(datum)
                    dfw.close()

                # read data in binary from file
                reader = open(FILENAME, 'rb')
                datum_reader = io.DatumReader()
                dfr = datafile.DataFileReader(reader, datum_reader)
                appended_data = []
                for datum in dfr:
                    appended_data.append(datum)

                print('Appended Data: %s' % appended_data)
                print('Appended Data Length: %d' % len(appended_data))
                is_correct = [datum] * 10 == appended_data
                if is_correct:
                    correct += 1
                print('Correct Appended: %s' % is_correct)
                print('')
        os.remove(FILENAME)
        self.assertEquals(correct, len(CODECS_TO_VALIDATE)*len(SCHEMAS_TO_VALIDATE))

    def test_context_manager(self):
        # Context manager was introduced as a first class
        # member only in Python 2.6 and above.
        import sys
        if sys.version_info < (2,6):
            print('Skipping context manager tests on this Python version.')
            return
        # Test the writer with a 'with' statement.
        writer = open(FILENAME, 'wb')
        datum_writer = io.DatumWriter()
        sample_schema, sample_datum = SCHEMAS_TO_VALIDATE[1]
        schema_object = schema.parse(sample_schema)
        with datafile.DataFileWriter(writer, datum_writer, schema_object) as dfw:
            dfw.append(sample_datum)
        self.assertTrue(writer.closed)

        # Test the reader with a 'with' statement.
        datums = []
        reader = open(FILENAME, 'rb')
        datum_reader = io.DatumReader()
        with datafile.DataFileReader(reader, datum_reader) as dfr:
            for datum in dfr:
                datums.append(datum)
        self.assertTrue(reader.closed)

    def test_metadata(self):
        # Test the writer with a 'with' statement.
        writer = open(FILENAME, 'wb')
        datum_writer = io.DatumWriter()
        sample_schema, sample_datum = SCHEMAS_TO_VALIDATE[1]
        schema_object = schema.parse(sample_schema)
        with datafile.DataFileWriter(writer, datum_writer, schema_object) as dfw:
            dfw.set_meta('test.string', 'foo')
            dfw.set_meta('test.number', '1')
            dfw.append(sample_datum)
        self.assertTrue(writer.closed)

        # Test the reader with a 'with' statement.
        datums = []
        reader = open(FILENAME, 'rb')
        datum_reader = io.DatumReader()
        with datafile.DataFileReader(reader, datum_reader) as dfr:
            self.assertEquals('foo', dfr.get_meta('test.string'))
            self.assertEquals('1', dfr.get_meta('test.number'))
            for datum in dfr:
                datums.append(datum)
        self.assertTrue(reader.closed)

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2017 Pluralsight LLC

import unittest
from spavro.schema_resolve import resolve
from spavro.exceptions import SchemaResolutionException


class TestResolver(unittest.TestCase):
    pass


pass_cases = (
("simple_null", "null", "null", "null"),
("simple_int", "int", "int", "int"),
("simple_long", "long", "long", "long"),
("simple_float", "float", "float", "float"),
("promote_int_to_long", "int", "long", "int"),
("promote_int_to_double", "int", "double", "int"),
("promote_float_to_double", "float", "double", "float"),
("promote_long_to_double", "long", "double", "long"),
("record_upgrade_to_union_and_default_field", {"fields": [{"default": "FOO",
                                 "type": {"symbols": ["FOO", "BAR"],
                                          "namespace": "",
                                          "type": "enum",
                                          "name": "F"},
                                 "name": "H"}
                                 ],
                     "type": "record",
                     "name": "Test"},
    ["int",
     {"fields": [{"default": "FOO",
                                     "type": {"symbols": ["FOO", "BAR"],
                                              "namespace": "",
                                              "type": "enum",
                                              "name": "F"},
                                     "name": "H"},
                                     {"name": "spork",
                                      "type": "int",
                                      "default": 1234}
                                     ],
                         "type": "record",
                         "name": "Test"}],
    {'fields': [{"type": {"symbols": ["FOO", "BAR"],
                                              "type": "enum",
                                              "name": "F"},
                                     "name": "H"},
     {'type': {'type': 'default', 'value': 1234},
      'name': 'spork'}],
  'type': 'record',
  'name': 'Test'}),
("symbol_added_to_reader_enum",
 {"type": "enum", "name": "bigby", "symbols": ["A", "C"]},
 {"type": "enum", "name": "bigby", "symbols": ["A", "B", "C"]},
 {'symbols': ['A', 'C'], 'type': 'enum', 'name': 'bigby'}),
("array_items_upgraded_to_union",
    {"type": "array", "items": "string"},
    {"type": "array", "items": ["int", "string"]},
    {'items': 'string', 'type': 'array'})
)

exception_cases = (
("null_vs_int", "null", "int", SchemaResolutionException),
("boolean_vs_int", "boolean", "int", SchemaResolutionException),
("lower_precision_promote_long_int", "long", "int", SchemaResolutionException),
("lower_precision_promote_double_float", "double", "float", SchemaResolutionException),
("missing_symbol_in_read",
 {"type": "enum", "name": "bigby", "symbols": ["A", "C"]},
 {"type": "enum", "name": "bigby", "symbols": ["A", "B"]}, SchemaResolutionException),
("union_missing_write_schema",
    "int", ["string", "boolean"], SchemaResolutionException),
("record_names_dont_match",
    {"type": "record", "name": "my_name", "fields": [{"type": "int", "name": "A"}]},
    {"type": "record", "name": "not_my_name", "fields": [{"type": "int", "name": "A"}]},
    SchemaResolutionException),
("record_field_types_dont_match",
    {"type": "record", "name": "my_name", "fields": [{"type": "string", "name": "A"}]},
    {"type": "record", "name": "my_name", "fields": [{"type": "int", "name": "A"}]},
    SchemaResolutionException),
("record_new_field_no_default",
    {"type": "record", "name": "my_name", "fields": [{"type": "string", "name": "A"}]},
    {"type": "record", "name": "my_name", "fields": [{"type": "int", "name": "A"},
                                                     {"type": "int", "name": "B"}]},
    SchemaResolutionException)
)


def create_pass_case(writer, reader, expected):
    def resolve_write_reader(self):
        resolved = resolve(writer, reader)
        self.assertEqual(resolved, expected)
    return resolve_write_reader


def create_exception_case(writer, reader, exception):
    def resolve_write_reader(self):
        with self.assertRaises(exception) as context:
            resolved = resolve(writer, reader)
    return resolve_write_reader


def make_cases(cases):
    for name, writer, reader, expected in cases:
        test_method = create_pass_case(writer, reader, expected)
        test_method.__name__ = 'test_schema_resolution_{}'.format(name)
        setattr(TestResolver, test_method.__name__, test_method)


def make_exception_cases(cases):
    for name, writer, reader, expected in cases:
        test_method = create_exception_case(writer, reader, expected)
        test_method.__name__ = 'test_incompatible_schema_{}'.format(name)
        setattr(TestResolver, test_method.__name__, test_method)


make_cases(pass_cases)
make_exception_cases(exception_cases)



# Modifications copyright (C) 2017 Pluralsight LLC
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
There are currently no IPC tests within python, in part because there are no
servers yet available.
"""
import unittest

# This test does import this code, to make sure it at least passes
# compilation.
from spavro import ipc

class TestIPC(unittest.TestCase):
  def test_placeholder(self):
    pass

  def test_server_with_path(self):
    client_with_custom_path = ipc.HTTPTransceiver('apache.org', 80, '/service/article')
    self.assertEqual('/service/article', client_with_custom_path.req_resource)

    client_with_default_path = ipc.HTTPTransceiver('apache.org', 80)
    self.assertEqual('/', client_with_default_path.req_resource)

if __name__ == '__main__':
  unittest.main()



# Copyright (C) 2017 Pluralsight LLC

import unittest
import six
from six import BytesIO as StringIO
# try:
#   from cStringIO import StringIO
# except ImportError:
#   from StringIO import StringIO
try:
    type(unicode)
except NameError:
    unicode = str

import json
import spavro.schema
import spavro.io
# from spavro.fast_binary import make_union_writer, get_writer
from spavro.io import FastDatumWriter, SlowDatumWriter


class TestOldVsNew(unittest.TestCase):
    def setUp(self):
        self.fdw = FastDatumWriter()
        self.sdw = SlowDatumWriter()

cases = (
("null", "null", None),
("int", "int", 1234),
("long", "long", 1234567890),
("float", "float", 1234.123),
("double", "double", 1234.12345),
("boolean", "boolean", True),
("string", "string", unicode('adsfasdf09809dsf-=adsf')),
("a_fixed_record", {"type": "fixed", "size": 4, "name": "fixeddata"}, b'Obj\x01'),
("an_enum_record", {"type": "enum", "symbols": ["A", "B", "C"], "name": "enumdata"}, "A"),
("an_array", {"type": "array", "items": "int"}, [1, 2, 3, 1234, 4321]),
("a_map", {"type": "map", "values": "int"}, {"L1": 1, "L2": 2, "L3": 3, "L4": 4}),
("union_test_string", ["null", "int", "string"], unicode("whassssuuuup")),
("union_test_int", ["null", "int", "string"], 1234),
("union_test_null", ["null", "int", "string"], None),
("a_record",
    {"type": "record", "name": "recorddata", "fields": [{"name": "field1", "type": "long"},
                                                        {"name": "field2", "type": "string"}]},
                                                        {"field1": 1234, "field2": unicode("whassssuuuuup")}),
("recursive_defined_record",
{u'fields': [{u'type': [u'null', u'string', {u'fields': [{u'type': u'Lisp', u'name': u'car'},
                                                         {u'type': u'Lisp', u'name': u'cdr'}],
                        u'type': u'record', u'name': u'Cons'}],
             u'name': u'value'}],
 u'type': u'record',
 u'name': u'Lisp'},
{'value': {'car': {'value': 'head'}, 'cdr': {'value': None}}}),
("union_of_two_records_recordA",
    [{"type": "record", "name": "recorddata", "fields": [{"name": "fieldA1", "type": "long"},
                                                        {"name": "fieldA2", "type": "string"}]},
     {"type": "record", "name": "recorddata2", "fields": [{"name": "fieldB1", "type": {"type": "fixed", "name": "fixedbytes", "size": 4}},
                                                           {"name": "fieldB2", "type": "string"}]}],
 {"fieldA1": 1234, "fieldA2": unicode("whassssuuuuup")}),
("union_of_two_records_recordB",
    [{"type": "record", "name": "recorddata", "fields": [{"name": "fieldA1", "type": "long"},
                                                        {"name": "fieldA2", "type": "string"}]},
     {"type": "record", "name": "recorddata2", "fields": [{"name": "fieldB1", "type": {"type": "fixed", "name": "fixedbytes", "size": 4}},
                                                           {"name": "fieldB2", "type": "string"}]}],
 {"fieldB1": b'\x01\x02\x03\x04', "fieldB2": unicode("Nother Record")})
)


def create_case(schema, datum):
    def compare_old_and_new(self):
        fastbuff = StringIO()
        slowbuff = StringIO()
        fastencoder = spavro.io.FastBinaryEncoder(fastbuff)
        slowencoder = spavro.io.SlowBinaryEncoder(slowbuff)
        write_schema = spavro.schema.parse(json.dumps(schema))
        for i in range(10):
            self.fdw.write_data(write_schema, datum, fastencoder)
            self.sdw.write_data(write_schema, datum, slowencoder)
        self.assertEqual(fastbuff.getvalue(), slowbuff.getvalue())
    return compare_old_and_new


def make_cases():
    for name, schema, datum in cases:
        test_method = create_case(schema, datum)
        test_method.__name__ = 'test_write_old_vs_new_{}'.format(name)
        setattr(TestOldVsNew, test_method.__name__, test_method)

make_cases()



# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spavro import schema
import json


class AvroTypeException(schema.AvroException):
    """Raised when datum is not an example of schema."""
    def __init__(self, expected_schema, datum):
        pretty_expected = json.dumps(json.loads(str(expected_schema)), indent=2)
        fail_msg = "The datum %s is not an example of the schema %s"\
                             % (datum, pretty_expected)
        schema.AvroException.__init__(self, fail_msg)


class SchemaResolutionException(schema.AvroException):
    def __init__(self, fail_msg, writers_schema=None, readers_schema=None):
        if writers_schema:
            pretty_writers = json.dumps(json.loads(str(writers_schema)), indent=2)
            fail_msg += "\nWriter's Schema: %s" % pretty_writers
        if readers_schema:
            pretty_readers = json.dumps(json.loads(str(readers_schema)), indent=2)
            fail_msg += "\nReader's Schema: %s" % pretty_readers
        schema.AvroException.__init__(self, fail_msg)



# Modifications copyright (C) 2017 Pluralsight LLC

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Protocol implementation.
"""
import six
try:
    from hashlib import md5
except ImportError:
    from md5 import md5
try:
    import json
except ImportError:
    import simplejson as json
from spavro import schema

#
# Constants
#

# TODO(hammer): confirmed 'fixed' with Doug
VALID_TYPE_SCHEMA_TYPES = ('enum', 'record', 'error', 'fixed')

#
# Exceptions
#


class ProtocolParseException(schema.AvroException):
    pass

#
# Base Classes
#


class Protocol(object):
    """An application protocol."""
    def _parse_types(self, types, type_names):
        type_objects = []
        for type in types:
            type_object = schema.make_avsc_object(type, type_names)
            if type_object.type not in VALID_TYPE_SCHEMA_TYPES:
                fail_msg = 'Type %s not an enum, fixed, record, or error.' % type
                raise ProtocolParseException(fail_msg)
            type_objects.append(type_object)
        return type_objects

    def _parse_messages(self, messages, names):
        message_objects = {}
        for name, body in six.iteritems(messages):
            if name in message_objects:
                fail_msg = 'Message name "%s" repeated.' % name
                raise ProtocolParseException(fail_msg)
            elif not(hasattr(body, 'get') and callable(body.get)):
                fail_msg = 'Message name "%s" has non-object body %s.' % (name, body)
                raise ProtocolParseException(fail_msg)
            request = body.get('request')
            response = body.get('response')
            errors = body.get('errors')
            message_objects[name] = Message(name, request, response, errors, names)
        return message_objects

    def __init__(self, name, namespace=None, types=None, messages=None):
        # Ensure valid ctor args
        if not name:
            fail_msg = 'Protocols must have a non-empty name.'
            raise ProtocolParseException(fail_msg)
        elif not isinstance(name, six.string_types):
            fail_msg = 'The name property must be a string.'
            raise ProtocolParseException(fail_msg)
        elif namespace is not None and not isinstance(namespace, six.string_types):
            fail_msg = 'The namespace property must be a string.'
            raise ProtocolParseException(fail_msg)
        elif types is not None and not isinstance(types, list):
            fail_msg = 'The types property must be a list.'
            raise ProtocolParseException(fail_msg)
        elif (messages is not None and 
                    not(hasattr(messages, 'get') and callable(messages.get))):
            fail_msg = 'The messages property must be a JSON object.'
            raise ProtocolParseException(fail_msg)

        self._props = {}
        self.set_prop('name', name)
        type_names = schema.Names()
        if namespace is not None: 
            self.set_prop('namespace', namespace)
            type_names.default_namespace = namespace
        if types is not None:
            self.set_prop('types', self._parse_types(types, type_names))
        if messages is not None:
            self.set_prop('messages', self._parse_messages(messages, type_names))
        self._md5 = md5(str(self).encode('utf-8')).digest()

    # read-only properties
    name = property(lambda self: self.get_prop('name'))
    namespace = property(lambda self: self.get_prop('namespace'))
    fullname = property(lambda self: 
                                            schema.Name(self.name, self.namespace).fullname)
    types = property(lambda self: self.get_prop('types'))
    types_dict = property(lambda self: dict([(type.name, type)
                                                                                     for type in self.types]))
    messages = property(lambda self: self.get_prop('messages'))
    md5 = property(lambda self: self._md5)
    props = property(lambda self: self._props)

    # utility functions to manipulate properties dict
    def get_prop(self, key):
        return self.props.get(key)
    def set_prop(self, key, value):
        self.props[key] = value  

    def to_json(self):
        to_dump = {}
        to_dump['protocol'] = self.name
        names = schema.Names(default_namespace=self.namespace)
        if self.namespace: 
            to_dump['namespace'] = self.namespace
        if self.types:
            to_dump['types'] = [ t.to_json(names) for t in self.types ]
        if self.messages:
            messages_dict = {}
            for name, body in six.iteritems(self.messages):
                messages_dict[name] = body.to_json(names)
            to_dump['messages'] = messages_dict
        return to_dump

    def __str__(self):
        return json.dumps(self.to_json())

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class Message(object):
    """A Protocol message."""
    def _parse_request(self, request, names):
        if not isinstance(request, list):
            fail_msg = 'Request property not a list: %s' % request
            raise ProtocolParseException(fail_msg)
        return schema.RecordSchema(None, None, request, names, 'request')
    
    def _parse_response(self, response, names):
        if isinstance(response, six.string_types) and names.has_name(response, None):
            return names.get_name(response, None)
        else:
            return schema.make_avsc_object(response, names)

    def _parse_errors(self, errors, names):
        if not isinstance(errors, list):
            fail_msg = 'Errors property not a list: %s' % errors
            raise ProtocolParseException(fail_msg)
        errors_for_parsing = {'type': 'error_union', 'declared_errors': errors}
        return schema.make_avsc_object(errors_for_parsing, names)

    def __init__(self,  name, request, response, errors=None, names=None):
        self._name = name

        self._props = {}
        self.set_prop('request', self._parse_request(request, names))
        self.set_prop('response', self._parse_response(response, names))
        if errors is not None:
            self.set_prop('errors', self._parse_errors(errors, names))

    # read-only properties
    name = property(lambda self: self._name)
    request = property(lambda self: self.get_prop('request'))
    response = property(lambda self: self.get_prop('response'))
    errors = property(lambda self: self.get_prop('errors'))
    props = property(lambda self: self._props)

    # utility functions to manipulate properties dict
    def get_prop(self, key):
        return self.props.get(key)
    def set_prop(self, key, value):
        self.props[key] = value

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self, names=None):
        if names is None:
            names = schema.Names()
        to_dump = {}
        to_dump['request'] = self.request.to_json(names)
        to_dump['response'] = self.response.to_json(names)
        if self.errors:
            to_dump['errors'] = self.errors.to_json(names)
        return to_dump

    def __eq__(self, that):
        return self.name == that.name and self.props == that.props


def make_avpr_object(json_data):
    """Build Avro Protocol from data parsed out of JSON string."""
    if hasattr(json_data, 'get') and callable(json_data.get):
        name = json_data.get('protocol')
        namespace = json_data.get('namespace')
        types = json_data.get('types')
        messages = json_data.get('messages')
        return Protocol(name, namespace, types, messages)
    else:
        raise ProtocolParseException('Not a JSON object: %s' % json_data)


def parse(json_string):
    """Constructs the Protocol from the JSON text."""
    try:
        json_data = json.loads(json_string)
    except:
        raise ProtocolParseException('Error parsing JSON: %s' % json_string)

    # construct the Avro Protocol object
    return make_avpr_object(json_data)




import unittest
from six import BytesIO

import spavro.io
from spavro.io import FastDatumWriter
# from spavro.io import SlowDatumWriter as FastDatumWriter
from spavro.exceptions import AvroTypeException


valid_data = (
("int_at_the_upper_boundary", 2147483647, '"int"'),
("int_at_the_lower_boundary", -2147483648, '"int"'),
("long_at_the_upper_boundary", 9223372036854775807, '"long"'),
("long_at_the_lower_boundary", -9223372036854775808, '"long"'),
("interger_data_float_schema", 123, '"float"'),
# booleans are considered an integer type? fascinating
("boolean_data_float_schema", True, '"float"'),
("boolean_data_integer_schema", True, '"int"'),
("optional_field", {"value": 100}, '{"fields": [{"type": ["null", "string"], "name": "id"}, {"type": "int", "name": "value"}], "type": "record", "name": "test_schema"}'),
("fixed", b'\x01\x01\x01\x01\x01\x01\x01\x01', '{"name": "testfix", "type": "fixed", "size": 8}'),
("make_sure_null_term_doesnt_break", b'\x01\x01\x00\x01\x01\x01\x01\x01', '{"name": "testfix", "type": "fixed", "size": 8}'),
)

invalid_data = (
("missing_required_field_1", {"value": 100}, '{"fields": [{"type": "string", "name": "id"}, {"type": "int", "name": "value"}], "type": "record", "name": "test_schema"}'),
("missing_required_field_2", {"id": "bork"}, '{"fields": [{"type": "string", "name": "id"}, {"type": "int", "name": "value"}], "type": "record", "name": "test_schema"}'),
("string_data_long_schema", u'boom!', '"long"'),
("string_data_boolean_schema", u"boom!", '"boolean"'),
("int_data_boolean_schema", 123, '"boolean"'),
("float_data_int_schema", 123.456, '"long"'),
("null_data_string_schema", None, '"string"'),
("null_data_int_schema", None, '"int"'),
("null_data_boolean_schema", None, '"boolean"'),
("mismatch_fixed_data_fixed_schema", b'\x97', '{"name": "testfix", "type": "fixed", "size": 8}'),
("int_too_big", 2147483648, '"int"'),
("int_too_small", -2147483649, '"int"'),
("long_too_big", 9223372036854775808, '"long"'),
("long_too_small", -9223372036854775809, '"long"'),
("wrong_data_in_array", [1, u'B'], '{"type": "array", "items": "int"}'),
)


class TestValidData(unittest.TestCase):
    pass


def create_good_case(schema, datum):
    write_schema = spavro.schema.parse(schema)
    def test_write_good_data(self):
        fastbuff = BytesIO()
        fastencoder = spavro.io.FastBinaryEncoder(fastbuff)
        fdw = FastDatumWriter(write_schema)
        fdw.write(datum, fastencoder)
    return test_write_good_data


def create_exception_case(schema, datum):
    print(schema)
    write_schema = spavro.schema.parse(schema)
    def test_write_invalid_data(self):
        with self.assertRaises(AvroTypeException) as context:
            fastbuff = BytesIO()
            fastencoder = spavro.io.FastBinaryEncoder(fastbuff)
            fdw = FastDatumWriter(write_schema)
            fdw.write(datum, fastencoder)
        print(context.exception)
    return test_write_invalid_data


def make_good_cases(cases):
    for name, datum, schema in cases:
        test_method = create_good_case(schema, datum)
        test_method.__name__ = 'test_good_data_{}'.format(name)
        setattr(TestValidData, test_method.__name__, test_method)


def make_exception_cases(cases):
    for name, datum, schema in cases:
        test_method = create_exception_case(schema, datum)
        test_method.__name__ = 'test_invalid_data_{}'.format(name)
        setattr(TestValidData, test_method.__name__, test_method)


make_good_cases(valid_data)
make_exception_cases(invalid_data)
